"""
"""
from batoms.bondsetting import Setting
import numpy as np
from time import time

class Polyhedrasetting(Setting):
    """
    Polyhedrasetting object

    The Polyhedrasetting object store the polyhedra information.

    Parameters:

    label: str
        The label define the batoms object that a Setting belong to.

    """
    def __init__(self, label) -> None:
        Setting.__init__(self, label)
        self.name = 'bpolyhedra'
        self.set_default(self.species)
    def __setitem__(self, index, value):
        """
        Add polyhedra one by one
        """
        p = self.find(index)
        if p is None:
            p = self.collection.add()
        p.symbol = index
        p.name = index
        p.color = value[0]
        p.edgewidth = value[1]
    def set_default(self, species):
        """
        """
        for sp, data in species.items():
            self[sp] = [np.append(data['color'][:3], 0.3), 0.005]
    def add_polyhedras(self, polyhedras):
        species = {sp: self.species[sp] for sp in polyhedras}
        self.set_default(species)
    def remove_polyhedras(self, polyhedras):
        for name in polyhedras:
            i = self.collection.find(name)
            if i != -1:
                self.collection.remove(i)
    def __repr__(self) -> str:
        s = '-'*60 + '\n'
        s = 'Center                color         edgewidth \n'
        for p in self.collection:
            s += '{0:10s}   [{1:1.2f}  {2:1.2f}  {3:1.2f}  {4:1.2f}]   {5:1.3f} \n'.format(\
                p.symbol, p.color[0], p.color[1], p.color[2], p.color[3], p.edgewidth)
        s += '-'*60 + '\n'
        return s


def build_polyhedralists(atoms, bondlists, bondsetting, polyhedrasetting):
        """
        Two modes:
        (1) Search atoms bonded to kind
        polyhedra_dict: {'kind': ligands}
        """
        from scipy.spatial import ConvexHull
        from batoms.tools import get_polyhedra_kind
        tstart = time()
        if 'species' not in atoms.info:
            atoms.info['species'] = atoms.get_chemical_symbols()
        speciesarray = np.array(atoms.info['species'])
        positions = atoms.positions
        polyhedra_kinds = {}
        polyhedra_dict = {}
        for b in bondsetting:
            spi = b.symbol1
            spj = b.symbol2
            if b.polyhedra:
                if spi not in polyhedra_dict: polyhedra_dict[spi] = []
                polyhedra_dict[spi].append(spj)
        # loop center atoms
        npositions = positions[bondlists[:, 1]] + np.dot(bondlists[:, 2:5], atoms.cell)
        for spi, spjs in polyhedra_dict.items():
            polyhedra_kind = get_polyhedra_kind(color = polyhedrasetting[spi].color, 
                                edgewidth = polyhedrasetting[spi].edgewidth)
            indis = np.where(speciesarray == spi)[0]
            for indi in indis:
                vertices = []
                indjs = np.where(bondlists[:, 0] == indi)[0]
                for indj in indjs:
                    if speciesarray[bondlists[indj, 1]] in spjs:
                        vertices.append(npositions[indj])
                nverts = len(vertices)
                if nverts >= 4:
                    # search convex polyhedra
                    hull = ConvexHull(vertices)
                    face = hull.simplices
                    nverts = len(polyhedra_kind['vertices'])
                    face = face + nverts
                    edge = []
                    for f in face:
                        edge.append([f[0], f[1]])
                        edge.append([f[0], f[2]])
                        edge.append([f[1], f[2]])
                    polyhedra_kind['vertices'] = polyhedra_kind['vertices'] + list(vertices)
                    polyhedra_kind['edges'] = polyhedra_kind['edges'] + list(edge)
                    polyhedra_kind['faces'] = polyhedra_kind['faces'] + list(face)
                    # print('edge: ', edge)
                    for e in edge:
                        # print(e)
                        center = (polyhedra_kind['vertices'][e[0]] + polyhedra_kind['vertices'][e[1]])/2.0
                        vec = polyhedra_kind['vertices'][e[0]] - polyhedra_kind['vertices'][e[1]]
                        length = np.linalg.norm(vec)
                        nvec = vec/length
                        # print(center, nvec, length)
                        polyhedra_kind['edge_cylinder']['lengths'].append(length/2.0)
                        polyhedra_kind['edge_cylinder']['centers'].append(center)
                        polyhedra_kind['edge_cylinder']['normals'].append(nvec)
            if len(polyhedra_kind['vertices']) > 0:
                polyhedra_kinds[spi] = polyhedra_kind
        # print('build_polyhedralists: {0:10.2f} s'.format(time() - tstart))
        return polyhedra_kinds



import bpy

class Base():
    def __init__(self, name):
        self.name = name
    @property
    def coll(self):
        return self.get_coll()
    def get_coll(self):
        return bpy.data.collections[self.name]


from ase import Atoms
import numpy as np
from time import time


def search_boundary(atoms, boundary = [[0, 1], [0, 1], [0, 1]], skin = 3):
    """
    search atoms in the boundary

    Parameters:

    atoms: ASE Atoms object

    boundary: list

    skin: float
        Could be the maximum cutoff.

    Return:

    atoms_boundary: ASE Atoms object
        atoms inside the boudary, not include the core
    offsets_skin: numpy array
        atoms close to boundary with a skin distance. 
        Used to search bond outside the boundary.
        index and offset

    """
    from ase.cell import Cell
    from math import floor, ceil
    tstart = time()
    if isinstance(boundary, float):
        boundary = [[-boundary, 1 + boundary], [-boundary, 1+boundary], [-boundary, 1+boundary]]
    boundary = np.array(boundary)
    boundary_skin = boundary.copy()
    # skin to scaled distance in cell
    par = atoms.cell.cellpar()
    skin = np.array([skin/par[0], skin/par[1], skin/par[2]])
    # skin region
    boundary_skin[:, 0] += skin
    boundary_skin[:, 1] -= skin
    # find supercell
    f = np.floor(boundary)
    c = np.ceil(boundary)
    ib = np.array([f[:, 0], c[:, 1]]).astype(int)
    M = np.product(ib[1] - ib[0] + 1)
    positions = atoms.get_scaled_positions()
    n = len(positions)
    npositions = np.tile(positions, (M - 1,) + (1,) * (len(positions.shape) - 1))
    i0 = 0
    # index
    offsets = np.zeros((M*n, 4), dtype=int)
    ind0 = np.arange(n).reshape(-1, 1)
    symbols0 = atoms.get_chemical_symbols()
    species0 = atoms.info['species']
    symbols = []
    species = []
    # repeat the positions so that
    # it completely covers the boundary
    for m0 in range(ib[0, 0], ib[1, 0] + 1):
        for m1 in range(ib[0, 1], ib[1, 1] + 1):
            for m2 in range(ib[0, 2], ib[1, 2] + 1):
                if m0 == 0 and m1 == 0 and m2 == 0: continue
                i1 = i0 + n
                npositions[i0:i1] += (m0, m1, m2)
                offsets[i0:i1] = np.append(ind0, np.array([[m0, m1, m2]]*n), axis = 1)
                symbols.extend(symbols0)
                species.extend(species0)
                i0 = i1
    # boundary condition
    ind1 =  np.where((npositions[:, 0] > boundary[0][0]) & (npositions[:, 0] < boundary[0][1]) \
                & (npositions[:, 1] > boundary[1][0]) & (npositions[:, 1] < boundary[1][1]) \
                & (npositions[:, 2] > boundary[2][0]) & (npositions[:, 2] < boundary[2][1]))
    # build atoms inside the boundary
    ind1 = list(set(ind1[0]))
    npositions1 = npositions[ind1]
    npositions1 = np.dot(npositions1, atoms.cell)
    symbols = np.array(symbols)[ind1]
    species = np.array(species)[ind1]
    atoms_boundary = Atoms(symbols, npositions1)
    atoms_boundary.info['species'] = species
    # build atoms inside the skin region
    # could be the atoms from core, thus add core
    npositions = np.append(npositions, positions, axis = 0)
    ind2 = np.append(ind1, i0 + ind0).astype(int)
    offsets[i0:i0+n] = np.append(ind0, np.array([[0, 0, 0]]*n), axis = 1)
    # atoms not belong to the skin region
    ind3 =  np.where((npositions[:, 0] > boundary_skin[0][0]) & (npositions[:, 0] < boundary_skin[0][1]) \
                & (npositions[:, 1] > boundary_skin[1][0]) & (npositions[:, 1] < boundary_skin[1][1])  #\
                & (npositions[:, 2] > boundary_skin[2][0]) & (npositions[:, 2] < boundary_skin[2][1]))
    ind3 = list(set(ind3[0]))
    ind3 = list(set(ind2) - set(ind3))
    offsets_skin = offsets[ind3]
    # print('search boundary: {0:10.2f} s'.format(time() - tstart))
    return atoms_boundary, offsets_skin

def search_bond(positions0, offsets_skin, bondlists, boundary, recursive = False, previous = None):
    """
    Search bonded atoms of sp1 or sp2 recursively.

    Parameters:

    positions0: array
        positions of original atoms, the core atoms
    offsets_skin: array
        index and offset of atoms of sp1 or sp2
    bondlists: array
        bondlist of the core atoms
    boundary: array
    recursive: bool
        recursive or not, for search mode 1 or 2
    previous: array
        offsets_skin of previous search. 
        The index should be remove from the result of this search.
    
    Return:

    offset_new: array
        index and offset of atoms which bond to sp1 or sp2
    
    """
    neighbors = np.array([]).reshape(-1, 4)
    if len(offsets_skin) == 0: return neighbors
    # build bonded atoms using offset_skin and bondlists
    sites = set(offsets_skin[:, 0])
    for i in sites:
        ind = np.where(bondlists[:, 0] == i)[0]
        if len(ind) == 0: continue
        neighbor = bondlists[ind][:, 1:]
        sitei = np.where(offsets_skin[:, 0] == i)[0]
        for j in sitei:
            temp = neighbor.copy()
            temp[:, 1:4] = temp[:, 1:4] + offsets_skin[j][1:4]
            neighbors = np.append(neighbors, temp, axis = 0)
    # choose index by two pricinple
    # 1, outside the boundary
    # 2, not from preious index
    # rebuild the positions, and check boundary
    neighbors = neighbors.astype(int)
    npositions = positions0[neighbors[:, 0]] + neighbors[:, 1:4]
    index =  np.where((npositions[:, 0] > boundary[0][0]) & (npositions[:, 0] < boundary[0][1]) \
                & (npositions[:, 1] > boundary[1][0]) & (npositions[:, 1] < boundary[1][1]) \
                & (npositions[:, 2] > boundary[2][0]) & (npositions[:, 2] < boundary[2][1]))
    #
    mask = np.ones(npositions.shape[0], dtype=bool)
    mask[index] = False
    offset_new = neighbors[mask]
    # offset_new = np.unique(offset_new, axis=0)
    # remove previous index, and remove duplicate
    if previous is not None:
        __, indices = np.unique(np.concatenate([previous, offset_new]), return_index=True, axis=0)
        indices = indices[indices >= len(previous)] - len(previous)
        offset_new = offset_new[indices]
    if len(offset_new) == 0: return np.array([]).reshape(-1, 4)
    # recursive search
    if recursive:
        offset_new1 = search_bond(positions0, offset_new, bondlists, boundary, recursive = True, previous = offsets_skin)
        # save
        if offset_new1 is not None:
            offset_new = np.append(offset_new, offset_new1, axis = 0)
    else:
        return offset_new
    return offset_new


class Boundary:
    """
    Boundary object.
    Example:
      cl = Boundary(atoms, d, index)
      cl.build()
    """

    def __init__(self, atoms, boundary_list, rotate_atoms = False):
        self.atoms = atoms
        self.natoms = len(atoms)
        self.cell = atoms.cell
        self.boundary_list = boundary_list
        # self.d = d
        # self.index = index
        self.rotate_atoms = rotate_atoms
        #
    def build(self, ):
        for boundary in self.boundary_list:
            print(boundary)
            self.cut(**boundary)
        return self.atoms
    def cut(self, atoms = None, d = None, index = None, direction = 1):
        """
        """
        if not atoms:
            atoms = self.atoms
        cell = atoms.cell
        normal = self.get_plane(d, index, cell)
        print(normal, d)
        # a*x + b*y + c*z - d = 0
        mask = []
        natoms = len(atoms)
        for i in range(natoms):
            v = atoms[i].position.dot(normal) - d
            # print(v)
            if v*direction > 0:
                mask.append(i)
        # view(atoms[mask])
        del atoms[mask]
        self.atoms = atoms
        if self.rotate_atoms:
            atoms = self.rotate()
    def rotate(self, atoms = None, index = None):
        """
        rotate normal of plane to z axis
        """
        import scipy
        if not atoms:
            atoms = self.atoms
        if not index:
            index = self.index
        cell = atoms.cell
        normal, d = self.get_plane(self.d, self.index, cell)
        vec = np.cross([0.0000014159, 0.000001951, 1], index)
        vec = vec/np.linalg.norm(vec)
        ang = np.arccos(normal[2])
        vec = ang*vec
        r = scipy.spatial.transform.Rotation.from_rotvec(vec)
        if scipy.version.version >= '1.4':
            mat = r.as_matrix()
        else: 
            mat = r.as_dcm()
        # print(mat)
        atoms.positions = atoms.positions.dot(mat)
        # atoms.cell = atoms.cell.dot(mat)
        self.atoms = atoms
        return atoms
    def get_plane(self, d, index = None, cell = None):
        '''
        plane equation: three point and distance from origin
        return normal and distance
        # a*x + b*y + c*z - d = 0
        '''
        index = [1.0/(index[i] + 0.000001) for i in range(3)]
        index = np.array(index)
        index = index/np.linalg.norm(index)
        points = cell*index
        # print(points)
        # x 
        v1 = points[1] - points[0]
        v2 = points[2] - points[0]
        normal = np.cross(v1, v2)
        print(v1, v2, normal)
        normal = normal/np.linalg.norm(normal)
        a, b, c = normal
        # d = np.dot(normal, points[2])
        return normal    
def build_bondlists(atoms, cutoff):
    """
    The default bonds are stored in 'default_bonds'
    Get all pairs of bonding atoms
    remove_bonds
    """
    from neighborlist import neighbor_list
    if len(cutoff) == 0: return {}
    #
    tstart = time()
    nli, nlj, nlS = neighbor_list('ijS', atoms, cutoff, self_interaction=False)
    bondlists = np.append(np.array([nli, nlj], dtype=int).T, np.array(nlS, dtype=int), axis = 1)
    # print('build_bondlists: {0:10.2f} s'.format(time() - tstart))
    return bondlists


if __name__ == "__main__":
    from ase.io import read, write
    from ase.build import molecule, bulk
    from ase.visualize import view
    from ase import Atom, Atoms
    # atoms = bulk('Pt', cubic = True)
    # atoms = atoms*[6, 6, 6]
    # cl = Boundary(atoms, boundary_list = [{'d': 10.0, 'index': [2, 2, 1]}])
    # atoms = cl.build()
    # view(atoms)
    # atoms = bulk('Pt', cubic = True)
    # atoms.write('pt.in')
    # view(atoms)
    # atoms = atoms*[3, 3, 3]
    atoms = read('docs/source/_static/datas/tio2.cif')
    atoms.info['species'] = atoms.get_chemical_symbols()
    # atoms = read('docs/source/_static/datas/mof-5.cif')
    # atoms.positions -= atoms.get_center_of_mass()
    # positions1, positions2, offset2 = search_boundary(atoms.positions, atoms.cell, boundary=[[0., 1.], [0., 1.], [0., 1.]])
    #positions1, positions2, offset2 = search_boundary(atoms.positions, atoms.cell, boundary=[[-2.6, 1.6], [-2.6, 1.6], [-2.6, 1.6]])
    cutoff = {('Ti', 'O'): [0, 3.0]}
    bondlists = build_bondlists(atoms, cutoff)
    print(bondlists)
    skins = np.array([[1, 4, 1, 1]])
    bondsetting = [('Ti', 'O')]
    boundary = np.array([[-0.6, 1.6], [-0.6, 1.6], [-0.6, 1.6]])
    # bondlists, only search bond
    # skins, only search atoms
    skins = search_bond(atoms.positions, skins, bondlists, boundary)
    print(skins)
    # skin = Atoms('Au'*len(positions2), positions = positions2)
    # view(atoms + skin)
    # bondsetting = {('Ti', 'O'): [0, 2.5, False, False], ('O', 'O'): [0, 1.5, False, False]}
    # bondpairs_boundary = search_skin(atoms, bondsetting, boundary=[[-0.6, 1.6], [-0.6, 1.6], [-0.6, 1.6]])
    # print(bondpairs_boundary)



import pytest
from batoms.butils import removeAll
from batoms.batoms import Batoms
from batoms.bio import read
import numpy as np



def test_boundary():
    from batoms.butils import removeAll
    from batoms.batoms import Batoms
    removeAll()
    a = 3.96
    positions = [[0, 0, 0], [a/2, a/2, 0], [a/2, 0, a/2], [0, a/2, a/2]]
    pt = Batoms('pt', {'Pt': positions}, pbc = True, cell = (a, a, a))
    pt.boundary = 0.01
    atoms = pt.get_atoms_with_boundary()
    assert len(atoms) == 14


def test_bond_search():
    from batoms.butils import removeAll
    from batoms.batoms import Batoms
    from batoms.bio import read
    removeAll()
    tio2 = read('../../docs/source/_static/datas/tio2.cif')
    tio2.boundary = 0.01
    tio2.model_type = 2
    atoms = tio2.get_atoms_with_boundary()
    assert len(atoms) == 55




if __name__ == '__main__':
    test_boundary()
    test_bond_search()
    print('\n Bondsetting: All pass! \n')


"""
Use this file for VScode debug only.

"""
import bpy

from batoms.batoms import Batoms
from batoms.batom import Batom

bl_info = {
    "name": "Batoms",
    "author": "Xing Wang",
    "version": (1, 0),
    "blender": (2, 93, 0),
    "description": "Python module for drawing and rendering beautiful atoms and molecules using blender.",
    "warning": "",
}

# Register
classes = [    ]

def register():
    for cls in classes:
        bpy.utils.register_class(cls)
    
def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

if __name__ == "__main__":

    register()



import pytest
from batoms.batom import Batom
from batoms.batoms import Batoms
from batoms.butils import removeAll
import numpy as np


def test_batom():
    removeAll()
    h = Batom(label = 'h2o', species = 'H', positions = [[0, -0.76, -0.2], [0, 0.76, -0.2]])
    o = Batom(label = 'h2o', species = 'O', positions = [[0, 0, 0.40]])
    h2o = Batoms('h2o', [h, o])
    assert isinstance(h2o, Batoms)

def test_batoms():
    """
    """
    removeAll()
    h2o = Batoms('h2o', {'O': [[0, 0, 0.40]], 'H': [[0, -0.76, -0.2], [0, 0.76, -0.2]]})
    assert isinstance(h2o, Batoms)
    # properties
    h2o.cell = [3, 3, 3]
    h2o.pbc = True
    h2o.model_type = 1
    h2o.translate([0, 0, 2])
    #
    h2o_2 = h2o.copy('h2o_2')
    assert isinstance(h2o_2, Batoms)
    h2o_3 = Batoms(label='h2o_2')
    #
    # delete
    del h2o_3['H'][[0]]
    assert len(h2o_3['H']) == 1
    #
    h2o_3.replace('O', 'C', [0])
    #
    h2o['O'][0] = [0, 0, 2.0]
    index = [0, 1]
    h2o['H'][index][:, 0] += 2

def test_extend():
    from ase.build import molecule
    co = molecule('CO')
    co = Batoms(label = 'co', atoms = co)
    co.translate([0, 0, 2])
    h2o = molecule('H2O')
    h2o = Batoms(label = 'h2o', atoms = h2o)
    h2o.extend(co)


def test_canvas():
    from ase.build import fcc111
    from batoms.batoms import Batoms
    from batoms.butils import removeAll
    removeAll()
    atoms = fcc111('Pt', size = (4, 4, 4), vacuum=0)
    pt111 = Batoms(label = 'pt111', atoms = atoms)
    pt111.cell[2, 2] += 5
    pt111.render([1, 1, 1])

def test_batoms_animation():
    from ase.build import molecule
    removeAll()
    atoms = molecule('C2H6SO')
    images = []
    for i in range(20):
        temp = atoms.copy()
        temp.rotate(18*i, 'z')
        images.append(temp)
    c2h6so = Batoms(label = 'c2h6so', atoms = images)
    c2h6so.load_frames()
    # c2h6so.render(animation = True)

def test_cavity():
    from batoms.bio import read
    from batoms.butils import removeAll
    removeAll()
    mof = read('../../docs/source/_static/datas/mof-5.cif')
    mof.draw_cavity_sphere(9.0, boundary = [[0.2, 0.8], [0.2, 0.8], [0.2, 0.8]])
    mof.model_type = 2
    mof.draw_cell()
    mof.render.light_energy = 30
    mof.render.run([1, 0, 0], engine = 'eevee')

def test_get_angles():
    from ase.build import molecule
    atoms = molecule('H2O')
    h2o = Batoms(atoms = atoms, label = 'h2o')
    angle = h2o.get_angle('H', 0, 'O', 0, 'H', 1)
    d = h2o.get_distances('H', 0, 'H', 1)



if __name__ == '__main__':
    test_batoms()
    test_batoms_animation()
    test_cavity()
    test_get_angles()
    print('\n Batoms: All pass! \n')


"""Definition of the Batom class.

This module defines the Batom object in the batoms package.

"""

from operator import pos
from os import name
from time import time
import bpy
import bmesh
from batoms.butils import object_mode
from batoms.data import material_styles_dict
from batoms.tools import get_atom_kind
import numpy as np

subcollections = ['atom', 'bond', 'instancer', 'instancer_atom', 'polyhedra', 'isosurface', 'text']

class Vertice():
    """Vertice Class
    
    Parameters:

    species: list of str
        The atomic structure built using ASE

    positions: array

    Examples:

    >>> from batoms.batom import Batom
    >>> c = Batom('C', [[0, 0, 0], [1.2, 0, 0]])
    >>> c.draw_atom()

    """
    

    def __init__(self, 
                species = None,
                position = None,
                element = None,
                x = None,
                y = None,
                z = None,
                ):
        #
        self.species = species
        self.position = position
        self.element = element
        self.x = x
        self.y = y
        self.z = z


class Batom():
    """Batom Class
    
    Then, a Batom object is linked to this main collection in Blender. 

    Parameters:

    label: str
        Name of the batom
    species: str
        species of the atoms.
    positions: array
        positions
    locations: array
        The object’s origin location in global coordinates.
    element: str
        element of the atoms
    color_style: str
        "JMOL", "ASE", "VESTA"

    Examples:

    >>> from batoms.batom import Batom
    >>> c = Batom('C', [[0, 0, 0], [1.2, 0, 0]])
    >>> c.draw_atom()

    """
    

    def __init__(self, 
                label = None,
                species = None,
                positions = None,
                location = np.array([0, 0, 0]),
                element = None,
                scale = 1.0, 
                segments = [32, 16],
                shape = 'UV_SPHERE',
                subdivisions = 2,
                props = {},
                color = None,
                transmit = None,
                color_style = 'JMOL',
                material_style = 'default',
                material = None,
                bsdf_inputs = None,
                 ):
        #
        self.scene = bpy.context.scene
        if species:
            self.label = label
            self.species = species
            if not element:
                self.element = species.split('_')[0]
            else:
                self.element = element
            self.name = 'atom_%s_%s'%(self.label, self.species)
            self.props = props
            self.color_style = color_style
            self.material_style = material_style
            self.bsdf_inputs = bsdf_inputs
            self.species_data = get_atom_kind(self.element, scale = scale, props = self.props, color_style = self.color_style)
            if color:
                self.species_data['color'] = color
            if transmit:
                self.species_data['color'] = transmit
            self.set_material(material)
            self.set_instancer(segments = segments, subdivisions = subdivisions, shape = shape)
            self.set_object(positions, location)
            self.batom.batom.radius = self.species_data['radius']
        else:
            self.from_batom(label)
    def set_material(self, material = None):
        name = 'material_atom_{0}_{1}'.format(self.label, self.species)
        if material:
            material = material.copy()
            material.name = name
        elif name not in bpy.data.materials:
            if not self.bsdf_inputs:
                bsdf_inputs = material_styles_dict[self.material_style]
            material = bpy.data.materials.new(name)
            material.diffuse_color = np.append(self.species_data['color'], self.species_data['transmit'])
            material.metallic = bsdf_inputs['Metallic']
            material.roughness = bsdf_inputs['Roughness']
            material.blend_method = 'BLEND'
            material.show_transparent_back = False
            material.use_nodes = True
            principled_node = material.node_tree.nodes['Principled BSDF']
            principled_node.inputs['Base Color'].default_value = np.append(self.species_data['color'], self.species_data['transmit'])
            principled_node.inputs['Alpha'].default_value = self.species_data['transmit']
            for key, value in bsdf_inputs.items():
                principled_node.inputs[key].default_value = value
    def object_mode(self):
        for object in bpy.data.objects:
            if object.mode == 'EDIT':
                bpy.ops.object.mode_set(mode = 'OBJECT')
    def set_instancer(self, segments = [32, 16], subdivisions = 2, shape = 'UV_SPHERE', shade_smooth = True):
        object_mode()
        name = 'instancer_atom_{0}_{1}'.format(self.label, self.species)
        if name not in bpy.data.objects:
            if shape.upper() == 'UV_SPHERE':
                bpy.ops.mesh.primitive_uv_sphere_add(segments = segments[0], ring_count = segments[1], radius = self.species_data['radius']) #, segments=32, ring_count=16)
            if shape.upper() == 'ICO_SPHERE':
                shade_smooth = False
                bpy.ops.mesh.primitive_ico_sphere_add(subdivisions = subdivisions, radius = self.species_data['radius']) #, segments=32, ring_count=16)
            if shape.upper() == 'CUBE':
                bpy.ops.mesh.primitive_cube_add(size = self.species_data['radius']) #, segments=32, ring_count=16)
                shade_smooth = False
            obj = bpy.context.view_layer.objects.active
            if isinstance(self.species_data['scale'], float):
                self.species_data['scale'] = [self.species_data['scale']]*3
            obj.scale = self.species_data['scale']
            obj.name = 'instancer_atom_{0}_{1}'.format(self.label, self.species)
            obj.data.materials.append(self.material)
            obj.data.name = 'instancer_atom_{0}_{1}'.format(self.label, self.species)
            if shade_smooth:
                bpy.ops.object.shade_smooth()
            obj.hide_set(True)
    def set_object(self, positions, location):
        """
        build child object and add it to main objects.
        """
        if self.name not in bpy.data.objects:
            mesh = bpy.data.meshes.new(self.name)
            obj_atom = bpy.data.objects.new(self.name, mesh)
            obj_atom.data.from_pydata(positions, [], [])
            obj_atom.location = location
            obj_atom.batom.is_batom = True
            obj_atom.batom.species = self.species
            obj_atom.batom.element = self.element
            obj_atom.batom.label = self.label
            bpy.data.collections['Collection'].objects.link(obj_atom)
        elif bpy.data.objects[self.name].batom.is_batom:
            obj_atom = bpy.data.objects[self.name]
        else:
            raise Exception("Failed, the name %s already in use and is not Batom object!"%self.name)
        self.instancer.parent = obj_atom
        obj_atom.instance_type = 'VERTS'
        bpy.context.view_layer.update()
    def from_batom(self, label):
        if label not in bpy.data.objects:
            raise Exception("%s is not a object!"%label)
        elif not bpy.data.objects[label].batom.is_batom:
            raise Exception("%s is not Batom object!"%label)
        ba = bpy.data.objects[label]
        self.species = ba.batom.species
        self.label = ba.batom.label
        self.element = ba.batom.element
    @property
    def batom(self):
        return self.get_batom()
    def get_batom(self):
        return bpy.data.objects['atom_%s_%s'%(self.label, self.species)]
    @property
    def instancer(self):
        return self.get_instancer()
    def get_instancer(self):
        return bpy.data.objects['instancer_atom_%s_%s'%(self.label, self.species)]
    @property
    def material(self):
        return self.get_material()
    def get_material(self):
        return bpy.data.materials['material_atom_%s_%s'%(self.label, self.species)]
    @property
    def scale(self):
        return self.get_scale()
    @scale.setter
    def scale(self, scale):
        self.set_scale(scale)
    def get_scale(self):
        return np.array(self.instancer.scale)
    def set_scale(self, scale):
        if isinstance(scale, float) or isinstance(scale, int):
            scale = [scale]*3
        self.instancer.scale = scale
    @property
    def radius(self):
        return self.get_radius()
    def get_radius(self):
        return np.array(self.batom.batom.radius)
    @property
    def size(self):
        return self.get_size()
    @size.setter
    def size(self, size):
        self.set_size(size)
    def get_size(self):
        return np.array(self.instancer.scale*self.radius)
    def set_size(self, size):
        scale = size/self.radius
        self.scale = [scale]*3
    @property
    def location(self):
        return self.get_location()
    def get_location(self):
        return np.array(self.batom.location)
    @property
    def local_positions(self):
        return self.get_local_positions()
    def get_local_positions(self):
        """
        using foreach_get and foreach_set to improve performance.
        """
        n = len(self)
        positions = np.empty(n*3, dtype=np.float64)
        self.batom.data.vertices.foreach_get('co', positions)  
        positions = positions.reshape((n, 3))
        return positions
    @property
    def positions(self):
        return self.get_positions()
    @positions.setter
    def positions(self, positions):
        self.set_positions(positions)
    def get_positions(self):
        """
        Get array of positions.
        """
        local_positions = self.local_positions
        n = len(local_positions)
        local_positions = np.append(local_positions, np.ones((n, 1)), axis = 1)
        mat= np.array(self.batom.matrix_world)
        positions = mat.dot(local_positions.T).T
        return positions[:, :3]
    def set_positions(self, positions):
        """
        Set positions
        """
        natom = len(self)
        if len(positions) != natom:
            raise ValueError('positions has wrong shape %s != %s.' %
                                (len(positions), natom))
        positions = positions - np.array(self.batom.location)
        positions = positions.reshape((natom*3, 1))
        self.batom.data.vertices.foreach_set('co', positions)
        self.batom.data.update()
    def get_scaled_positions(self, cell):
        """
        Get array of scaled_positions.
        """
        from ase.cell import Cell
        cell = Cell.new(cell)
        scaled_positions = cell.scaled_positions(self.local_positions)
        return scaled_positions
        
    @property
    def color(self):
        return self.get_color()
    @color.setter
    def color(self, color):
        """
        >>> h.color = [0.8, 0.1, 0.3, 1.0]
        """
        self.set_color(color)
    def get_color(self):
        Viewpoint_color = self.material.diffuse_color
        BSDF_color = self.bsdf['Base bsdf'].default_value
        Alpha = self.bsdf['Alpha'].default_value
        return Viewpoint_color, BSDF_color, Alpha
    def set_color(self, color):
        if len(color) == 3:
            color = [color[0], color[1], color[2], 1]
        self.material.diffuse_color = color
        self.material.node_tree.nodes['Principled BSDF'].inputs['Base Color'].default_value = color
        self.material.node_tree.nodes['Principled BSDF'].inputs['Alpha'].default_value = color[3]
    @property
    def bsdf(self):
        return self.get_bsdf()
    @bsdf.setter
    def bsdf(self, bsdf):
        self.set_bsdf(bsdf)
    def get_bsdf(self):
        return self.material.node_tree.nodes['Principled BSDF'].inputs
    def set_bsdf(self, bsdf):
        for key, value in bsdf.items():
            self.material.node_tree.nodes['Principled BSDF'].inputs[key].default_value = value
    @property
    def segments(self):
        return self.get_segments()
    @segments.setter
    def segments(self, segments):
        self.set_segments(segments)
    def get_segments(self):
        nverts = len(self.instancer.data.vertices)
        return nverts
    def set_segments(self, segments):
        if not isinstance(segments, int):
            raise Exception('Segments should be int!')
        self.clean_batoms_objects('instancer_atom_%s_%s'%(self.label, self.species))
        self.set_instancer(segments = segments)
        self.instancer.parent = self.batom
    @property
    def subdivisions(self):
        return self.get_subdivisions()
    @subdivisions.setter
    def subdivisions(self, subdivisions):
        self.set_subdivisions(subdivisions)
    def get_subdivisions(self):
        nverts = len(self.instancer.data.vertices)
        return nverts
    def set_subdivisions(self, subdivisions):
        if not isinstance(subdivisions, int):
            raise Exception('subdivisions should be int!')
        self.clean_batoms_objects('instancer_atom_%s_%s'%(self.label, self.species))
        self.set_instancer(subdivisions = subdivisions, shape='ICO_SPHERE')
        self.instancer.parent = self.batom
    @property
    def shape(self):
        return self.get_shape()
    @shape.setter
    def shape(self, shape):
        self.set_shape(shape)
    def get_shape(self):
        """
        todo"""
        # nverts = len(self.instancer.data.vertices)
        return 'to do'
    def set_shape(self, shape):
        if shape.upper() not in ["UV_SPHERE", "ICO_SPHERE", "CUBE"]:
            raise Exception('Shape %s is not supported!'%shape)
        self.clean_batoms_objects('instancer_atom_%s_%s'%(self.label, self.species))
        self.set_instancer(shape = shape)
        self.instancer.parent = self.batom
    def clean_batoms_objects(self, obj):
        obj = bpy.data.objects[obj]
        bpy.data.objects.remove(obj, do_unlink = True)
    def delete_verts(self, index = []):
        """
        delete verts
        """
        object_mode()
        obj = self.batom
        bm = bmesh.new()
        bm.from_mesh(obj.data)
        bm.verts.ensure_lookup_table()
        verts_select = [bm.verts[i] for i in index] 
        bmesh.ops.delete(bm, geom=verts_select, context='VERTS')
        if len(bm.verts) == 0:
            bpy.data.objects.remove(obj)
            bpy.data.objects.remove(self.instancer)
        else:
            bm.to_mesh(obj.data)
    def delete(self, index = []):
        """
        delete atom.

        index: list
            index of atoms to be delete
        
        For example, delete the second atom in h species. 
        Please note that index start from 0.

        >>> h.delete([1])

        """
        if isinstance(index[0], (bool, np.bool_)):
            index = np.where(index)[0]
        if isinstance(index, int):
            index = [index]
        self.delete_verts(index)
    def __delitem__(self, index):
        """
        """
        self.delete(index)
    def draw_constraints(self):
        """
        """
        #
        constr = self.atoms.constraints
        self.constrainatom = []
        for c in constr:
            if isinstance(c, FixAtoms):
                for n, i in enumerate(c.index):
                    self.constrainatom += [i]
    
    def load_frames(self, images = []):
        """

        images: list
            list of positions
        
        >>> from batoms import Batom
        >>> import numpy as np
        >>> positions = np.array([[0, 0 ,0], [1.52, 0, 0]])
        >>> h = Batom('h2o', 'H', positions)
        >>> images = []
        >>> for i in range(10):
                images.append(positions + [0, 0, i])
        >>> h.load_frames(images)
        
        """
        # render settings
        nimage = len(images)
        batom = self.batom
        nverts = len(batom.data.vertices)
        for i in range(0, nimage):
            positions = images[i]
            for j in range(nverts):
                batom.data.vertices[j].co = np.array(positions[j]) - np.array(batom.location)
                batom.data.vertices[j].keyframe_insert('co', frame=i + 1)
        self.scene.frame_start = 1
        self.scene.frame_end = nimage
    
    def __len__(self):
        return len(self.batom.data.vertices)
    
    def __getitem__(self, index):
        """Return a subset of the Batom.

        i -- int, describing which atom to return.

        #todo: this is slow for large system
        
        """
        return self.positions[index]
        # batom = self.batom
        # if isinstance(index, int):
        #     natom = len(self)
        #     if index < -natom or index >= natom:
        #         raise IndexError('Index out of range.')
        #     return batom.matrix_world @ batom.data.vertices[index].co
        # if isinstance(index, list):
        #     positions = np.array([self[i] for i in index])
        #     return positions
        # if isinstance(index, slice):
        #     start, stop, step = index.indices(len(self))
        #     index = list(range(start, stop, step))
        #     return self[index]
        # if isinstance(index, tuple):
        #     i, j = index
        #     positions = self[i]
        #     return positions[:, j]

    def __setitem__(self, index, value):
        """Return a subset of the Batom.

        i -- int, describing which atom to return.

        #todo: this is slow for large system

        """
        positions = self.positions
        positions[index] = value
        self.set_positions(positions)

        # batom  =self.batom
        # if isinstance(index, int):
        #     natom = len(self)
        #     if index < -natom or index >= natom:
        #         raise IndexError('Index out of range.')
        #     batom.data.vertices[index].co = np.array(value) - np.array(batom.location)
        # if isinstance(index, list):
        #     for i in index:
        #         self[i] = value[i]
        # if isinstance(index, tuple):
        #     i, j = index
        #     batom.data.vertices[i].co[j] = np.array(value) - np.array(batom.location[j])

    def repeat(self, m, cell):
        """
        In-place repeat of atoms.

        >>> from batoms.batom import Batom
        >>> c = Batom('co', 'C', [[0, 0, 0], [1.2, 0, 0]])
        >>> c.repeat([3, 3, 3], np.array([[5, 0, 0], [0, 5, 0], [0, 0, 5]]))
        """
        if isinstance(m, int):
            m = (m, m, m)
        for x, vec in zip(m, cell):
            if x != 1 and not vec.any():
                raise ValueError('Cannot repeat along undefined lattice '
                                 'vector')
        M = np.product(m)
        n = len(self)
        positions = np.tile(self.local_positions, (M,) + (1,) * (len(self.local_positions.shape) - 1))
        i0 = 0
        for m0 in range(m[0]):
            for m1 in range(m[1]):
                for m2 in range(m[2]):
                    i1 = i0 + n
                    positions[i0:i1] += np.dot((m0, m1, m2), cell)
                    i0 = i1
        self.add_vertices(positions[n:])
    def copy(self, label, species):
        """
        Return a copy.

        name: str
            The name of the copy.

        For example, copy H species:
        
        >>> h_new = h.copy(label = 'h_new', species = 'H')

        """
        object_mode()
        batom = Batom(label, species, self.local_positions, location = self.batom.location, scale = self.scale, material=self.material)
        return batom
    def extend(self, other):
        """
        Extend batom object by appending batom from *other*.
        
        >>> from batoms.batoms import Batom
        >>> h1 = Batom('h2o', 'H_1', [[0, 0, 0], [2, 0, 0]])
        >>> h2 = Batom('h2o', 'H_2', [[0, 0, 2], [2, 0, 2]])
        >>> h = h1 + h2
        """
        # could also use self.add_vertices(other.positions)
        object_mode()
        bpy.ops.object.select_all(action='DESELECT')
        self.batom.select_set(True)
        other.batom.select_set(True)
        bpy.context.view_layer.objects.active = self.batom
        bpy.ops.object.join()
    def __iadd__(self, other):
        """
        >>> h1 += h2
        """
        self.extend(other)
        return self
    def __add__(self, other):
        """
        >>> h1 = h1 + h2
        """
        self += other
        return self
    def __iter__(self):
        batom = self.batom
        for i in range(len(self)):
            yield batom.matrix_world @ batom.data.vertices[i].co
    def __repr__(self):
        s = "Batom('%s', positions = %s" % (self.species, list(self.positions))
        return s
    def add_vertices(self, positions):
        """
        """
        object_mode()
        bm = bmesh.new()
        bm.from_mesh(self.batom.data)
        bm.verts.ensure_lookup_table()
        verts = []
        for pos in positions:
            bm.verts.new(pos)
        bm.to_mesh(self.batom.data)
    def translate(self, displacement):
        """Translate atomic positions.

        The displacement argument is an xyz vector.

        For example, move H species molecule by a vector [0, 0, 5]

        >>> h.translate([0, 0, 5])
        """
        object_mode()
        bpy.ops.object.select_all(action='DESELECT')
        self.batom.select_set(True)
        bpy.ops.transform.translate(value=displacement)
    def rotate(self, angle, axis = 'Z', orient_type = 'GLOBAL'):
        """Rotate atomic based on a axis and an angle.

        Parameters:

        angle: float
            Angle that the atoms is rotated around the axis.
        axis: str
            'X', 'Y' or 'Z'.

        For example, rotate h2o molecule 90 degree around 'Z' axis:
        
        >>> h.rotate(90, 'Z')

        """
        object_mode()
        bpy.ops.object.select_all(action='DESELECT')
        self.batom.select_set(True)
        bpy.ops.transform.rotate(value=angle, orient_axis=axis.upper(), orient_type = orient_type)
    def get_cell(self):
        if not self.label in bpy.data.collections:
            return None
        bcell = bpy.data.collections['%s_cell'%self.label]
        cell = np.array([bcell.matrix_world @ bcell.data.vertices[i].co for i in range(3)])
        return cell


import bpy
from ase import Atom, Atoms



def object_mode():
    for object in bpy.data.objects:
            if object.mode == 'EDIT':
                bpy.ops.object.mode_set(mode = 'OBJECT')

def read_batoms_collection_list():
    """
    Read all batoms collection 
    """
    items = [col.name for col in bpy.data.collections if col.is_batoms]
    return items
def read_atoms_list(coll):
    '''   
    '''
    coll_atom_kinds = [coll for coll in coll.children if 'atoms' in coll.name][0]
    elements = []
    for obj in coll_atom_kinds.all_objects:
        ele = obj.name.split('_')[2]
        elements.append(ele)
    return elements
        

def read_atoms_select():
    '''   
    '''
    from ase import Atoms, Atom
    atoms = Atoms()
    cell_vertexs = []
    for obj in bpy.context.selected_objects:
        if "BOND" in obj.name.upper():
            continue
        if obj.type not in {'MESH', 'SURFACE', 'META'}:
            continue
        name = ""
        if 'atom_kind_' == obj.name[0:10]:
            print(obj.name)
            ind = obj.name.index('atom_kind_')
            ele = obj.name[ind + 10:].split('_')[0]
            if len(obj.children) != 0:
                for vertex in obj.data.vertices:
                    location = obj.matrix_world @ vertex.co
                    atoms.append(Atom(ele, location))
            else:
                if not obj.parent:
                    location = obj.location
                    atoms.append(Atom(ele, location))
        # cell
        if 'point_cell' == obj.name[0:10]:
            print(obj.name)
            if len(obj.children) != 0:
                for vertex in obj.data.vertices:
                    location = obj.matrix_world @ vertex.co
                    # print(location)
                    cell_vertexs.append(location)
    # print(atoms)
    # print(cell_vertexs)
    if cell_vertexs:
        cell = [cell_vertexs[4], cell_vertexs[2], cell_vertexs[1]]
        atoms.cell = cell
        atoms.pbc = [True, True, True]
    # self.atoms = atoms
    return atoms

def remove_collection(name):
        collection = bpy.data.collections.get(name)
        for obj in collection.all_objects:
            bpy.data.objects.remove(obj, do_unlink=True)
        for coll in collection.children:
            bpy.data.collections.remove(coll)
        bpy.data.collections.remove(collection)
        
def clean_objects():
    for item in bpy.data.objects:
        bpy.data.objects.remove(item)

def removeAll():
    #types =  ['MESH', 'CURVE', 'SURFACE', 'META', 'FONT', 'ARMATURE', 'LATTICE', 'EMPTY', 'GPENCIL', 'CAMERA', 'LIGHT', 'SPEAKER', 'LIGHT_PROBE']
    for mesh in bpy.data.meshes:
        bpy.data.meshes.remove(mesh)
    for obj in bpy.data.objects:
        bpy.data.objects.remove(obj)
    for cam in bpy.data.cameras:
        bpy.data.cameras.remove(cam)
    for light in bpy.data.lights:
        bpy.data.lights.remove(light)
    for mat in bpy.data.materials:
        bpy.data.materials.remove(mat)
    for coll in bpy.data.collections:
        if coll.name == 'Collection': continue
        bpy.data.collections.remove(coll)

def lock_camera_to_view(switch):
    for area in bpy.context.screen.areas:
        if area.type == 'VIEW_3D':
            for space in area.spaces:
                if space.type == 'VIEW_3D':
                    space.lock_camera = switch


import pytest
from batoms.cell import Bcell
import numpy as np

def test_bcell():
    """
    """
    cell = Bcell(label = 'pt', array = [2, 2, 2])
    assert isinstance(cell, Bcell)
    # properties
    cell[2, 2] += 5
    assert cell[2, 2] == 7
    # repeat
    cell.repeat([2, 2, 2])
    assert np.allclose(cell.array, np.array([[4, 0, 0], [0, 4, 0], [0, 0, 14]]))
    cell2 = cell.copy('pt-2')
    assert isinstance(cell2, Bcell)

if __name__ == '__main__':
    test_bcell()
    print('\n Bcell: All pass! \n')


"""
"""
from operator import add
import bpy
import numpy as np
from ase.cell import Cell
from batoms.butils import object_mode

class Bcell():
    """
    Unit cell of three dimensions.

    """
    def __init__(self, label, array = np.zeros([3, 3]), location = np.array([0, 0, 0])) -> None:
        """
        ver: 3x3 verlike object
          The three cell vectors: cell[0], cell[1], and cell[2].
        """
        self.label = label
        self.name = 'cell_%s_edge'%(self.label)
        self.draw_cell_edge(array, location)
    def draw_cell_edge(self, array, location):
        """
        Draw unit cell by edge, however, can not be rendered.
        """
        if array is None:
            array = np.zeros([3, 3])
        if not len(array) == 8:
            cell = Cell.new(array)
            verts = self.array2verts(cell.array)
        else:
            verts = array - array[3]
            location = array[3]
        if self.name not in bpy.data.objects:
            edges = [[3, 0], [3, 1], [4, 0], [4, 1],
                    [2, 5], [2, 6], [7, 5], [7, 6], 
                    [3, 2], [0, 6], [1, 5], [4, 7]
            ]
            mesh = bpy.data.meshes.new("cell_%s_edge"%self.label)
            mesh.from_pydata(verts, edges, [])  
            mesh.update()
            for f in mesh.polygons:
                f.use_smooth = True
            obj_edge = bpy.data.objects.new("cell_%s_edge"%self.label, mesh)
            obj_edge.data = mesh
            obj_edge.location = location
            obj_edge.bcell.is_bcell = True
            bpy.data.collections['Collection'].objects.link(obj_edge)
        elif bpy.data.objects[self.name].bcell.is_bcell:
            print('%s exist and is bcell, use it.'%self.name)
        else:
            raise Exception("Failed, the name %s already in use and is not Bcell object!"%self.name)
        bpy.context.view_layer.update()
    def __repr__(self) -> str:
        numbers = self.array.tolist()
        s = 'Cell({})'.format(numbers)
        return s
    def __getitem__(self, index):
        return self.array[index]
    def __setitem__(self, index, value):
        """
        Add bondpair one by one
        """
        """Set unit cell vectors.

        Parameters:

        Examples:

        """
        from ase.cell import Cell
        from ase.geometry.cell import complete_cell
        bcell = self.bcell
        array = self.array
        array[index] = value
        verts = self.array2verts(array)
        for i in range(8):
            bcell.data.vertices[i].co = np.array(verts[i])
    def __array__(self, dtype=float):
        if dtype != float:
            raise ValueError('Cannot convert cell to array of type {}'
                             .format(dtype))
        return self.array
    @property
    def bcell(self):
        return self.get_bcell()
    def get_bcell(self):
        return bpy.data.objects['cell_%s_edge'%(self.label)]
    @property
    def array(self):
        return self.get_array()
    def get_array(self):
        cell = np.array([self.local_verts[0] - self.local_verts[3],
                         self.local_verts[1] - self.local_verts[3],
                         self.local_verts[2] - self.local_verts[3]])
        return cell
    @property
    def local_verts(self):
        return self.get_local_verts()
    def get_local_verts(self):
        bcell = self.bcell
        return np.array([bcell.data.vertices[i].co for i in range(8)])
    @property
    def verts(self):
        return self.get_verts()
    def get_verts(self):
        return np.array([self.bcell.matrix_world @ self.bcell.data.vertices[i].co for i in range(8)])
    def array2verts(self, array):
        """
        """
        verts = np.array([[1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
            [0, 0, 0],
            [1, 1, 0],
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 1],
            ])
        verts = np.dot(verts, array)
        return verts
    @property
    def location(self):
        return self.get_location()
    def get_location(self):
        return np.array(self.bcell.location)
    def copy(self, label):
        object_mode()
        cell = Bcell(label, array = self.array, location = self.bcell.location)
        return cell
    def repeat(self, m):
        self[:] = np.array([m[c] * self.array[c] for c in range(3)])
    


import bpy
import numpy as np
from batoms.data import material_styles_dict
import time
#========================================================
def draw_cell_curve(coll, verts, label = None):
    """
    Draw unit cell by edge, however, can not be rendered.
    """
    if verts is not None:
        edges = [0, 4, 6, 2, 0, 1, 5, 7, 3, 1]
        crv = bpy.data.curves.new("edge_cell", 'CURVE')
        crv.dimensions = '3D'
        spline = crv.splines.new(type='NURBS')
        spline.points.add(len(edges)-1)
        for p, i in zip(spline.points, edges):
            p.co = np.append(verts[i], [1.0]) # (add nurbs weight)
        cell = bpy.data.objects.new("cell_%s_edge"%label, crv)
        coll.objects.link(cell)
def draw_cell_cylinder(coll_cell, cell_vertices, label = None, celllinewidth = 0.05):
    """
    Draw unit cell using cylinder.
    """
    if cell_vertices is not None:
        # build materials
        material = bpy.data.materials.new('cell_{0}'.format(label))
        # material.label = 'cell'
        material.diffuse_color = (0.0, 0.0, 0.0, 1.0)
        # draw points
        bpy.ops.mesh.primitive_uv_sphere_add(radius = celllinewidth) #, segments=32, ring_count=16)
        sphere = bpy.context.view_layer.objects.active
        sphere.name = 'instancer_cell_%s_sphere'%label
        sphere.data.materials.append(material)
        bpy.ops.object.shade_smooth()
        sphere.hide_set(True)
        mesh = bpy.data.meshes.new('point_cell' )
        obj_cell = bpy.data.objects.new('cell_%s_point'%label, mesh )
        # Associate the vertices
        obj_cell.data.from_pydata(cell_vertices, [], [])
        sphere.parent = obj_cell
        obj_cell.instance_type = 'VERTS'
        coll_cell.objects.link(sphere)
        coll_cell.objects.link(obj_cell)
        #
        # edges
        edges = [[3, 0], [3, 1], [4, 0], [4, 1],
                    [2, 5], [2, 6], [7, 5], [7, 6], 
                    [3, 2], [0, 6], [1, 5], [4, 7]
            ]
        cell_edges = {'lengths': [], 
                      'centers': [],
                      'normals': []}
        for e in edges:
            center = (cell_vertices[e[0]] + cell_vertices[e[1]])/2.0
            vec = cell_vertices[e[0]] - cell_vertices[e[1]]
            length = np.linalg.norm(vec)
            nvec = vec/length
            # print(center, nvec, length)
            cell_edges['lengths'].append(length)
            cell_edges['centers'].append(center)
            cell_edges['normals'].append(nvec)
        #
        source = bond_source(vertices=6)
        verts, faces = cylinder_mesh_from_instance_vec(cell_edges['centers'], cell_edges['normals'], cell_edges['lengths'], celllinewidth, source)
        # print(verts)
        mesh = bpy.data.meshes.new("cell_cylinder")
        mesh.from_pydata(verts, [], faces)  
        mesh.update()
        for f in mesh.polygons:
            f.use_smooth = True
        obj_edge = bpy.data.objects.new("cell_%s_cylinder"%label, mesh)
        obj_edge.data = mesh
        obj_edge.data.materials.append(material)
        bpy.ops.object.shade_smooth()
        coll_cell.objects.link(obj_edge)


def draw_text(coll_text = None, atoms = None, type = None):
    tstart = time.time()
    positions = atoms.positions
    n = len(positions)
    for i in range(n):
        location = positions[i] + np.array([0, 0, 1.0])
        FontCurve = bpy.data.curves.new(type="FONT",name="myFontCurve")
        ob = bpy.data.objects.new("myFontOb",FontCurve)
        if type == 0:
            ob.data.body = "%s"%i
        elif type == 1:
            ob.data.body = "%s"%atoms[i].symbol
        ob.location = location
        coll_text.objects.link(ob)
    print('text: {0:10.2f} s'.format(time.time() - tstart))


def draw_bond_kind(kind, 
                   datas, 
                   label = None,
                   coll = None,
                   source = None, 
                   bsdf_inputs = None, 
                   material_style = 'plastic'):
    if len(datas['centers']) == 0:
        return
    if not bsdf_inputs:
        bsdf_inputs = material_styles_dict[material_style]
    if datas['style'] in ['0', '1']:
        vertices = 16
    elif datas['style'] in ['2', '3']:
        vertices = 6
    source = bond_source(vertices = vertices)
    tstart = time.time()
    material = bpy.data.materials.new('bond_kind_{0}'.format(kind))
    material.diffuse_color = np.append(datas['color'], datas['transmit'])
    material.metallic = bsdf_inputs['Metallic']
    material.roughness = bsdf_inputs['Roughness']
    material.blend_method = 'BLEND'
    material.use_nodes = True
    principled_node = material.node_tree.nodes['Principled BSDF']
    principled_node.inputs['Base Color'].default_value = np.append(datas['color'], datas['transmit'])
    principled_node.inputs['Alpha'].default_value = datas['transmit']
    for key, value in bsdf_inputs.items():
        principled_node.inputs[key].default_value = value
    datas['materials'] = material
    #
    verts, faces = cylinder_mesh_from_instance_vec(datas['centers'], datas['normals'], datas['lengths'], datas['width'], source)
    mesh = bpy.data.meshes.new("mesh_kind_{0}".format(kind))
    mesh.from_pydata(verts, [], faces)  
    mesh.update()
    for f in mesh.polygons:
        f.use_smooth = True
    obj_bond = bpy.data.objects.new("bond_{0}_{1}".format(label, kind), mesh)
    obj_bond.data = mesh
    obj_bond.data.materials.append(material)
    bpy.ops.object.shade_smooth()
    coll.objects.link(obj_bond)
    # print('bonds: {0}   {1:10.2f} s'.format(kind, time.time() - tstart))
    
def draw_cavity(kind, 
                   datas, 
                   label = None,
                   coll = None,
                   source = None, 
                   bsdf_inputs = None, 
                   material_style = 'plastic'):
    if len(datas['edges']) == 0:
        return
    if not bsdf_inputs:
        bsdf_inputs = material_styles_dict[material_style]
    tstart = time.time()
    material = bpy.data.materials.new('bond_kind_{0}'.format(kind))
    material.diffuse_color = np.append(datas['color'], datas['transmit'])
    material.metallic = bsdf_inputs['Metallic']
    material.roughness = bsdf_inputs['Roughness']
    material.blend_method = 'BLEND'
    material.use_nodes = True
    principled_node = material.node_tree.nodes['Principled BSDF']
    principled_node.inputs['Base Color'].default_value = np.append(datas['color'], datas['transmit'])
    principled_node.inputs['Alpha'].default_value = datas['transmit']
    for key, value in bsdf_inputs.items():
        principled_node.inputs[key].default_value = value
    datas['materials'] = material
    #
    mesh = bpy.data.meshes.new("mesh_kind_{0}".format(kind))
    mesh.from_pydata(datas['vertices'], datas['edges'], [])
    mesh.update()
    name = "cavity_{0}_{1}".format(label, kind)
    obj_bond = bpy.data.objects.new(name, mesh)
    obj_bond.data = mesh
    obj_bond.data.materials.append(material)
    coll.objects.link(obj_bond)
    bpy.context.view_layer.objects.active = bpy.context.scene.objects.get(name)
    bpy.ops.object.mode_set(mode = 'EDIT')
    # fill edge with face
    bpy.ops.mesh.edge_face_add()
    bpy.ops.object.mode_set(mode = 'OBJECT')
    bpy.ops.object.shade_smooth()
    # print('bonds: {0}   {1:10.2f} s'.format(kind, time.time() - tstart))
    


def draw_polyhedra_kind(kind, 
                        datas, 
                        label = None,
                        coll = None,
                        source = None, 
                        show_edge = True,
                        bsdf_inputs = None, 
                        material_style = 'default'):
        """
        """
        if not bsdf_inputs:
            bsdf_inputs = material_styles_dict[material_style]
        tstart = time.time()
        material = bpy.data.materials.new('polyhedra_kind_{0}'.format(kind))
        material.diffuse_color = np.append(datas['color'], datas['transmit'])
        material.blend_method = 'BLEND'
        material.use_nodes = True
        principled_node = material.node_tree.nodes['Principled BSDF']
        principled_node.inputs['Base Color'].default_value = np.append(datas['color'], datas['transmit'])
        principled_node.inputs['Alpha'].default_value = datas['transmit']
        for key, value in bsdf_inputs.items():
            principled_node.inputs[key].default_value = value
        datas['materials'] = material
        #
        # create new mesh structure
        mesh = bpy.data.meshes.new("mesh_kind_{0}".format(kind))
        # mesh.from_pydata(datas['vertices'], datas['edges'], datas['faces'])  
        mesh.from_pydata(datas['vertices'], [], datas['faces'])  
        mesh.update()
        for f in mesh.polygons:
            f.use_smooth = False
        obj_polyhedra = bpy.data.objects.new("polyhedra_{0}_{1}_face".format(label, kind), mesh)
        obj_polyhedra.data = mesh
        obj_polyhedra.data.materials.append(material)
        # bpy.ops.object.shade_smooth()
        # bpy.ops.object.shade_flat()
        coll.objects.link(obj_polyhedra)
        #---------------------------------------------------
        if not show_edge: return
        source = bond_source(vertices=6)
        material = bpy.data.materials.new('polyhedra_edge_kind_{0}'.format(kind))
        material.diffuse_color = np.append(datas['edge_cylinder']['color'], datas['edge_cylinder']['transmit'])
        # material.blend_method = 'BLEND'
        material.use_nodes = True
        principled_node = material.node_tree.nodes['Principled BSDF']
        principled_node.inputs['Base Color'].default_value = np.append(datas['edge_cylinder']['color'], datas['edge_cylinder']['transmit'])
        principled_node.inputs['Alpha'].default_value = datas['transmit']
        for key, value in bsdf_inputs.items():
            principled_node.inputs[key].default_value = value
        verts, faces = cylinder_mesh_from_instance_vec(datas['edge_cylinder']['centers'], datas['edge_cylinder']['normals'], datas['edge_cylinder']['lengths'], datas['edgewidth'], source)
        # print(verts)
        mesh = bpy.data.meshes.new("mesh_kind_{0}".format(kind))
        mesh.from_pydata(verts, [], faces)  
        mesh.update()
        for f in mesh.polygons:
            f.use_smooth = True
        obj_edge = bpy.data.objects.new("polyhedra_{0}_{1}_edge".format(label, kind), mesh)
        obj_edge.data = mesh
        obj_edge.data.materials.append(material)
        bpy.ops.object.shade_smooth()
        # STRUCTURE.append(obj_polyhedra)
        coll.objects.link(obj_edge)
        # print('polyhedras: {0}   {1:10.2f} s'.format(kind, time.time() - tstart))


def draw_isosurface(coll_isosurface, verts, faces, color,
                    bsdf_inputs = None, material_style = 'default'):
    """Computes an isosurface from a volume grid.
    
    Parameters:

    
    """
    #material
    if not bsdf_inputs:
        bsdf_inputs = material_styles_dict[material_style]
    material = bpy.data.materials.new('isosurface')
    material.name = 'isosurface'
    material.diffuse_color = color
    material.blend_method = 'BLEND'
    material.use_nodes = True
    principled_node = material.node_tree.nodes['Principled BSDF']
    principled_node.inputs['Base Color'].default_value = color
    principled_node.inputs['Alpha'].default_value = color[3]
    for key, value in bsdf_inputs.items():
            principled_node.inputs[key].default_value = value
    # create new mesh structure
    isosurface = bpy.data.meshes.new("isosurface")
    isosurface.from_pydata(verts, [], faces)  
    isosurface.update()
    for f in isosurface.polygons:
        f.use_smooth = True
    iso_object = bpy.data.objects.new("isosurface", isosurface)
    iso_object.data = isosurface
    iso_object.data.materials.append(material)
    bpy.ops.object.shade_smooth()
    coll_isosurface.objects.link(iso_object)


# draw bonds
def bond_source(vertices = 12, depth = 1.0):
    bpy.ops.mesh.primitive_cylinder_add(vertices = vertices, depth = depth)
    cyli = bpy.context.view_layer.objects.active
    me = cyli.data
    verts = []
    faces = []
    for vertices in me.vertices:
        verts.append(np.array(vertices.co))
    for poly in me.polygons:
        face = []
        for loop_index in range(poly.loop_start, poly.loop_start + poly.loop_total):
            # print("    Vertex: %d" % me.loops[loop_index].vertex_index)
            face.append(me.loops[loop_index].vertex_index)
        faces.append(face)
    cyli.select_set(True)
    bpy.data.objects.remove(cyli, do_unlink = True)
    n = len(faces[0])
    faces1 = [faces[i] for i in range(len(faces)) if len(faces[i]) == n]
    faces2 = [faces[i] for i in range(len(faces)) if len(faces[i]) != n]
    return verts, faces1, faces2
# draw atoms
def atom_source():
    bpy.ops.mesh.primitive_uv_sphere_add() #, segments=32, ring_count=16)
    # bpy.ops.mesh.primitive_cylinder_add()
    sphe = bpy.context.view_layer.objects.active
    me = sphe.data
    verts = []
    faces = []
    for vertices in me.vertices:
        verts.append(np.array(vertices.co))
    for poly in me.polygons:
        face = []
        for loop_index in range(poly.loop_start, poly.loop_start + poly.loop_total):
            # print("    Vertex: %d" % me.loops[loop_index].vertex_index)
            face.append(me.loops[loop_index].vertex_index)
        faces.append(face)
    sphe.select_set(True)
    bpy.ops.object.delete()
    return [verts, faces]



def sphere_mesh_from_instance(centers, radius, source):
    verts = []
    faces = []
    vert0, face0 = source
    nvert = len(vert0)
    nb = len(centers)
    for i in range(nb):
        center = centers[i]
        # r = radius[i]
        # normal = normal/np.linalg.norm(normal)
        for vert in vert0:
            vert = vert*[radius, radius, radius]
            vert += center
            verts.append(vert)
        for face in face0:
            face = [x+ i*nvert for x in face]
            faces.append(face)
    return verts, faces

def cylinder_mesh_from_instance_vec(centers, normals, lengths, scale, source):
    from scipy.spatial.transform import Rotation as R
    tstart = time.time()
    vert0, face1, face2 = source
    nb = len(centers)
    nvert = len(vert0)
    centers = np.array(centers)
    normals = np.array(normals)
    vert0 = np.array(vert0)
    face1 = np.array(face1)
    face2 = np.array(face2)
    scale = np.array([[scale, scale]]*len(centers))
    scale = np.append(scale, np.array(lengths).reshape(-1, 1), axis = 1)
    # print(np.cross([0.0000014159, 0.000001951, 1], normals[0]))
    vec = np.cross([0.0, 0.0, 1], normals) + np.array([0.000000001, 0, 0])
    vec = vec/np.linalg.norm(vec, axis = 1)[:, None]
    # print(np.arccos(normals[0, 2]*0.999999))
    ang = np.arccos(normals[:, 2]*0.999999)
    # print(-1*ang[0]*vec[0])
    vec = -1*(vec.T*ang).T
    # print(R.from_rotvec(vec[0]).as_matrix())
    r = R.from_rotvec(vec)
    matrix = r.as_matrix()
    # print('vert1 0: ', vert0*scale[0])
    verts = np.tile(vert0, (nb, 1, 1))
    verts = verts*scale[:, None]
    # print('matrix: ', verts[0].dot(matrix[0]))
    verts = np.matmul(verts, matrix)
    # print('matrix: ', verts)
    # print(verts.shape)
    # centers = np.tile(centers, (nb, 1))
    # print('center: ', verts[0] + centers[0])
    verts += centers[:, None]
    # print('center: ', verts)
    # faces.append(face)
    verts = verts.reshape(-1, 3)
    #----------------------------
    nf1 = len(face1[0])
    nf2 = len(face2[0])
    faces1 = np.tile(face1, (nb, 1, 1))
    faces2 = np.tile(face2, (nb, 1, 1))
    offset = np.arange(nb)*nvert
    offset = offset.reshape(-1, 1, 1)
    faces1 = faces1 + offset
    faces1 = faces1.reshape(-1, nf1)
    faces2 = faces2 + offset
    faces2 = faces2.reshape(-1, nf2)
    faces = list(faces1) + list(face2)
    # print('cylinder_mesh_from_instance: {0:10.2f} s'.format( time.time() - tstart))
    return verts, faces

def draw_plane(location = (0, 0, -0.01), color = (0.2, 0.2, 1.0, 1.0), size = 200, bsdf_inputs = None, material_style = 'default'):
    """
    Draw a plane.
    location: array
    color: array
    size: float
    bsdf_inputs: dict
    material_style: str
    """
    # build materials
    if not bsdf_inputs:
        bsdf_inputs = material_styles_dict[material_style]
    material = bpy.data.materials.new('plane')
    material.name = 'plane'
    material.diffuse_color = color
    # material.blend_method = 'BLEND'
    material.use_nodes = True
    principled_node = material.node_tree.nodes['Principled BSDF']
    principled_node.inputs['Alpha'].default_value = color[3]
    for key, value in bsdf_inputs.items():
            principled_node.inputs[key].default_value = value
    # Instantiate a floor plane
    bpy.ops.mesh.primitive_plane_add(size=size, location=location)
    current_object = bpy.context.object
    current_object.data.materials.append(material)


"""
From ASE
https://wiki.fysik.dtu.dk/ase/_modules/ase/neighborlist.html#neighbor_list

Modified to support:

1) species instead of element
2) add a bondlengt range
    e.g cutoff = {('Ti_1', 'O'):[0.5, 3.0] }

"""

import numpy as np

from ase.data import atomic_numbers
from ase.geometry import complete_cell


import time

def primitive_neighbor_list(quantities, pbc, cell, positions, cutoff,
                            species=None, self_interaction=False,
                            use_scaled_positions=False, max_nbins=1e6):
    """Compute a neighbor list for an atomic configuration.

    Atoms outside periodic boundaries are mapped into the box. Atoms
    outside nonperiodic boundaries are included in the neighbor list
    but complexity of neighbor list search for those can become n^2.

    The neighbor list is sorted by first atom index 'i', but not by second
    atom index 'j'.

    Parameters:

    quantities: str
        Quantities to compute by the neighbor list algorithm. Each character
        in this string defines a quantity. They are returned in a tuple of
        the same order. Possible quantities are

            * 'i' : first atom index
            * 'j' : second atom index
            * 'd' : absolute distance
            * 'D' : distance vector
            * 'S' : shift vector (number of cell boundaries crossed by the bond
              between atom i and j). With the shift vector S, the
              distances D between atoms can be computed from:
              D = positions[j]-positions[i]+S.dot(cell)
    pbc: array_like
        3-tuple indicating giving periodic boundaries in the three Cartesian
        directions.
    cell: 3x3 matrix
        Unit cell vectors.
    positions: list of xyz-positions
        Atomic positions.  Anything that can be converted to an ndarray of
        shape (n, 3) will do: [(x1,y1,z1), (x2,y2,z2), ...]. If
        use_scaled_positions is set to true, this must be scaled positions.
    cutoff: float or dict
        Cutoff for neighbor search. It can be:

            * A single float: This is a global cutoff for all elements.
            * A dictionary: This specifies cutoff values for element
              pairs. Specification accepts element numbers of symbols.
              Example: {(1, 6): 1.1, (1, 1): 1.0, ('C', 'C'): 1.85}
            * A list/array with a per atom value: This specifies the radius of
              an atomic sphere for each atoms. If spheres overlap, atoms are
              within each others neighborhood. See :func:`~ase.neighborlist.natural_cutoffs`
              for an example on how to get such a list.
    self_interaction: bool
        Return the atom itself as its own neighbor if set to true.
        Default: False
    use_scaled_positions: bool
        If set to true, positions are expected to be scaled positions.
    max_nbins: int
        Maximum number of bins used in neighbor search. This is used to limit
        the maximum amount of memory required by the neighbor list.

    Returns:

    i, j, ... : array
        Tuple with arrays for each quantity specified above. Indices in `i`
        are returned in ascending order 0..len(a)-1, but the order of (i,j)
        pairs is not guaranteed.

    """

    # Naming conventions: Suffixes indicate the dimension of an array. The
    # following convention is used here:
    #     c: Cartesian index, can have values 0, 1, 2
    #     i: Global atom index, can have values 0..len(a)-1
    #     xyz: Bin index, three values identifying x-, y- and z-component of a
    #          spatial bin that is used to make neighbor search O(n)
    #     b: Linearized version of the 'xyz' bin index
    #     a: Bin-local atom index, i.e. index identifying an atom *within* a
    #        bin
    #     p: Pair index, can have value 0 or 1
    #     n: (Linear) neighbor index

    # Return empty neighbor list if no atoms are passed here
    tstart = time.time()
    if len(positions) == 0:
        empty_types = dict(i=(int, (0, )),
                           j=(int, (0, )),
                           D=(float, (0, 3)),
                           d=(float, (0, )),
                           S=(int, (0, 3)))
        retvals = []
        for i in quantities:
            dtype, shape = empty_types[i]
            retvals += [np.array([], dtype=dtype).reshape(shape)]
        if len(retvals) == 1:
            return retvals[0]
        else:
            return tuple(retvals)

    # Compute reciprocal lattice vectors.
    b1_c, b2_c, b3_c = np.linalg.pinv(cell).T

    # Compute distances of cell faces.
    l1 = np.linalg.norm(b1_c)
    l2 = np.linalg.norm(b2_c)
    l3 = np.linalg.norm(b3_c)
    face_dist_c = np.array([1 / l1 if l1 > 0 else 1,
                            1 / l2 if l2 > 0 else 1,
                            1 / l3 if l3 > 0 else 1])

    if isinstance(cutoff, dict):
        max_cutoff = np.array([x for x in cutoff.values()]).max()
    else:
        if np.isscalar(cutoff):
            max_cutoff = cutoff
        else:
            cutoff = np.asarray(cutoff)
            max_cutoff = 2*np.max(cutoff)

    # We use a minimum bin size of 3 A
    bin_size = max(max_cutoff, 3)
    # Compute number of bins such that a sphere of radius cutoff fits into
    # eight neighboring bins.
    nbins_c = np.maximum((face_dist_c / bin_size).astype(int), [1, 1, 1])
    nbins = np.prod(nbins_c)
    # Make sure we limit the amount of memory used by the explicit bins.
    while nbins > max_nbins:
        nbins_c = np.maximum(nbins_c // 2, [1, 1, 1])
        nbins = np.prod(nbins_c)

    # Compute over how many bins we need to loop in the neighbor list search.
    neigh_search_x, neigh_search_y, neigh_search_z = \
        np.ceil(bin_size * nbins_c / face_dist_c).astype(int)

    # If we only have a single bin and the system is not periodic, then we
    # do not need to search neighboring bins
    neigh_search_x = 0 if nbins_c[0] == 1 and not pbc[0] else neigh_search_x
    neigh_search_y = 0 if nbins_c[1] == 1 and not pbc[1] else neigh_search_y
    neigh_search_z = 0 if nbins_c[2] == 1 and not pbc[2] else neigh_search_z

    # Sort atoms into bins.
    if use_scaled_positions:
        scaled_positions_ic = positions
        positions = np.dot(scaled_positions_ic, cell)
    else:
        scaled_positions_ic = np.linalg.solve(complete_cell(cell).T,
                                              positions.T).T
    bin_index_ic = np.floor(scaled_positions_ic*nbins_c).astype(int)
    cell_shift_ic = np.zeros_like(bin_index_ic)

    for c in range(3):
        if pbc[c]:
            # (Note: np.divmod does not exist in older numpies)
            cell_shift_ic[:, c], bin_index_ic[:, c] = \
                divmod(bin_index_ic[:, c], nbins_c[c])
        else:
            bin_index_ic[:, c] = np.clip(bin_index_ic[:, c], 0, nbins_c[c]-1)

    # Convert Cartesian bin index to unique scalar bin index.
    bin_index_i = (bin_index_ic[:, 0] +
                   nbins_c[0] * (bin_index_ic[:, 1] +
                                 nbins_c[1] * bin_index_ic[:, 2]))

    # atom_i contains atom index in new sort order.
    atom_i = np.argsort(bin_index_i)
    bin_index_i = bin_index_i[atom_i]

    # Find max number of atoms per bin
    max_natoms_per_bin = np.bincount(bin_index_i).max()

    # Sort atoms into bins: atoms_in_bin_ba contains for each bin (identified
    # by its scalar bin index) a list of atoms inside that bin. This list is
    # homogeneous, i.e. has the same size *max_natoms_per_bin* for all bins.
    # The list is padded with -1 values.
    atoms_in_bin_ba = -np.ones([nbins, max_natoms_per_bin], dtype=int)
    for i in range(max_natoms_per_bin):
        # Create a mask array that identifies the first atom of each bin.
        mask = np.append([True], bin_index_i[:-1] != bin_index_i[1:])
        # Assign all first atoms.
        atoms_in_bin_ba[bin_index_i[mask], i] = atom_i[mask]

        # Remove atoms that we just sorted into atoms_in_bin_ba. The next
        # "first" atom will be the second and so on.
        mask = np.logical_not(mask)
        atom_i = atom_i[mask]
        bin_index_i = bin_index_i[mask]

    # Make sure that all atoms have been sorted into bins.
    assert len(atom_i) == 0
    assert len(bin_index_i) == 0

    # Now we construct neighbor pairs by pairing up all atoms within a bin or
    # between bin and neighboring bin. atom_pairs_pn is a helper buffer that
    # contains all potential pairs of atoms between two bins, i.e. it is a list
    # of length max_natoms_per_bin**2.
    atom_pairs_pn = np.indices((max_natoms_per_bin, max_natoms_per_bin),
                               dtype=int)
    atom_pairs_pn = atom_pairs_pn.reshape(2, -1)

    # Initialized empty neighbor list buffers.
    first_at_neightuple_nn = []
    secnd_at_neightuple_nn = []
    cell_shift_vector_x_n = []
    cell_shift_vector_y_n = []
    cell_shift_vector_z_n = []

    # This is the main neighbor list search. We loop over neighboring bins and
    # then construct all possible pairs of atoms between two bins, assuming
    # that each bin contains exactly max_natoms_per_bin atoms. We then throw
    # out pairs involving pad atoms with atom index -1 below.
    binz_xyz, biny_xyz, binx_xyz = np.meshgrid(np.arange(nbins_c[2]),
                                               np.arange(nbins_c[1]),
                                               np.arange(nbins_c[0]),
                                               indexing='ij')
    # The memory layout of binx_xyz, biny_xyz, binz_xyz is such that computing
    # the respective bin index leads to a linearly increasing consecutive list.
    # The following assert statement succeeds:
    #     b_b = (binx_xyz + nbins_c[0] * (biny_xyz + nbins_c[1] *
    #                                     binz_xyz)).ravel()
    #     assert (b_b == np.arange(np.prod(nbins_c))).all()


    # print('neighborlist 0: {0:10.2f} s'.format( time.time() - tstart))
    tstart = time.time()

    # First atoms in pair.
    _first_at_neightuple_n = atoms_in_bin_ba[:, atom_pairs_pn[0]]
    for dz in range(-neigh_search_z, neigh_search_z+1):
        for dy in range(-neigh_search_y, neigh_search_y+1):
            for dx in range(-neigh_search_x, neigh_search_x+1):
                # Bin index of neighboring bin and shift vector.
                shiftx_xyz, neighbinx_xyz = divmod(binx_xyz + dx, nbins_c[0])
                shifty_xyz, neighbiny_xyz = divmod(biny_xyz + dy, nbins_c[1])
                shiftz_xyz, neighbinz_xyz = divmod(binz_xyz + dz, nbins_c[2])
                neighbin_b = (neighbinx_xyz + nbins_c[0] *
                              (neighbiny_xyz + nbins_c[1] * neighbinz_xyz)
                              ).ravel()

                # Second atom in pair.
                _secnd_at_neightuple_n = \
                    atoms_in_bin_ba[neighbin_b][:, atom_pairs_pn[1]]

                # Shift vectors.
                _cell_shift_vector_x_n = \
                    np.resize(shiftx_xyz.reshape(-1, 1),
                              (max_natoms_per_bin**2, shiftx_xyz.size)).T
                _cell_shift_vector_y_n = \
                    np.resize(shifty_xyz.reshape(-1, 1),
                              (max_natoms_per_bin**2, shifty_xyz.size)).T
                _cell_shift_vector_z_n = \
                    np.resize(shiftz_xyz.reshape(-1, 1),
                              (max_natoms_per_bin**2, shiftz_xyz.size)).T

                # We have created too many pairs because we assumed each bin
                # has exactly max_natoms_per_bin atoms. Remove all surperfluous
                # pairs. Those are pairs that involve an atom with index -1.
                mask = np.logical_and(_first_at_neightuple_n != -1,
                                      _secnd_at_neightuple_n != -1)
                if mask.sum() > 0:
                    first_at_neightuple_nn += [_first_at_neightuple_n[mask]]
                    secnd_at_neightuple_nn += [_secnd_at_neightuple_n[mask]]
                    cell_shift_vector_x_n += [_cell_shift_vector_x_n[mask]]
                    cell_shift_vector_y_n += [_cell_shift_vector_y_n[mask]]
                    cell_shift_vector_z_n += [_cell_shift_vector_z_n[mask]]

    # Flatten overall neighbor list.
    first_at_neightuple_n = np.concatenate(first_at_neightuple_nn)
    secnd_at_neightuple_n = np.concatenate(secnd_at_neightuple_nn)
    cell_shift_vector_n = np.transpose([np.concatenate(cell_shift_vector_x_n),
                                        np.concatenate(cell_shift_vector_y_n),
                                        np.concatenate(cell_shift_vector_z_n)])

    # Add global cell shift to shift vectors
    cell_shift_vector_n += cell_shift_ic[first_at_neightuple_n] - \
        cell_shift_ic[secnd_at_neightuple_n]

    # Remove all self-pairs that do not cross the cell boundary.
    if not self_interaction:
        m = np.logical_not(np.logical_and(
            first_at_neightuple_n == secnd_at_neightuple_n,
            (cell_shift_vector_n == 0).all(axis=1)))
        first_at_neightuple_n = first_at_neightuple_n[m]
        secnd_at_neightuple_n = secnd_at_neightuple_n[m]
        cell_shift_vector_n = cell_shift_vector_n[m]

    # For nonperiodic directions, remove any bonds that cross the domain
    # boundary.
    for c in range(3):
        if not pbc[c]:
            m = cell_shift_vector_n[:, c] == 0
            first_at_neightuple_n = first_at_neightuple_n[m]
            secnd_at_neightuple_n = secnd_at_neightuple_n[m]
            cell_shift_vector_n = cell_shift_vector_n[m]

    # Sort neighbor list.
    i = np.argsort(first_at_neightuple_n)
    first_at_neightuple_n = first_at_neightuple_n[i]
    secnd_at_neightuple_n = secnd_at_neightuple_n[i]
    cell_shift_vector_n = cell_shift_vector_n[i]

    # Compute distance vectors.
    distance_vector_nc = positions[secnd_at_neightuple_n] - \
        positions[first_at_neightuple_n] + \
        cell_shift_vector_n.dot(cell)
    abs_distance_vector_n = \
        np.sqrt(np.sum(distance_vector_nc*distance_vector_nc, axis=1))

    # We have still created too many pairs. Only keep those with distance
    # smaller than max_cutoff.
    mask = abs_distance_vector_n < max_cutoff
    first_at_neightuple_n = first_at_neightuple_n[mask]
    secnd_at_neightuple_n = secnd_at_neightuple_n[mask]
    cell_shift_vector_n = cell_shift_vector_n[mask]
    distance_vector_nc = distance_vector_nc[mask]
    abs_distance_vector_n = abs_distance_vector_n[mask]
    # print('neighborlist 1: {0:10.2f} s'.format( time.time() - tstart))

    tstart = time.time()
    if isinstance(cutoff, dict) and species is not None:
        # If cutoff is a dictionary, then the cutoff radii are specified per
        # element pair. We now have a list up to maximum cutoff.
        species = np.array(species)
        per_pair_cutoff_n = np.zeros_like(abs_distance_vector_n)
        per_pair_cutoff_n_min = np.zeros_like(abs_distance_vector_n)
        for (sp1, sp2), c in cutoff.items():
            mask = np.logical_and(
                    species[first_at_neightuple_n] == sp1,
                    species[secnd_at_neightuple_n] == sp2)
            per_pair_cutoff_n_min[mask] = c[0]
            per_pair_cutoff_n[mask] = c[1]
        mask1 = abs_distance_vector_n < per_pair_cutoff_n
        mask2 = abs_distance_vector_n > per_pair_cutoff_n_min
        mask = mask1 & mask2
        first_at_neightuple_n = first_at_neightuple_n[mask]
        secnd_at_neightuple_n = secnd_at_neightuple_n[mask]
        cell_shift_vector_n = cell_shift_vector_n[mask]
        distance_vector_nc = distance_vector_nc[mask]
        abs_distance_vector_n = abs_distance_vector_n[mask]
    elif not np.isscalar(cutoff):
        # If cutoff is neither a dictionary nor a scalar, then we assume it is
        # a list or numpy array that contains atomic radii. Atoms are neighbors
        # if their radii overlap.
        mask = abs_distance_vector_n < \
            cutoff[first_at_neightuple_n] + cutoff[secnd_at_neightuple_n]
        first_at_neightuple_n = first_at_neightuple_n[mask]
        secnd_at_neightuple_n = secnd_at_neightuple_n[mask]
        cell_shift_vector_n = cell_shift_vector_n[mask]
        distance_vector_nc = distance_vector_nc[mask]
        abs_distance_vector_n = abs_distance_vector_n[mask]
    # print('neighborlist 2: {0:10.2f} s'.format( time.time() - tstart))
    # Assemble return tuple.
    retvals = []
    for q in quantities:
        if q == 'i':
            retvals += [first_at_neightuple_n]
        elif q == 'j':
            retvals += [secnd_at_neightuple_n]
        elif q == 'D':
            retvals += [distance_vector_nc]
        elif q == 'd':
            retvals += [abs_distance_vector_n]
        elif q == 'S':
            retvals += [cell_shift_vector_n]
        else:
            raise ValueError('Unsupported quantity specified.')
    if len(retvals) == 1:
        return retvals[0]
    else:
        return tuple(retvals)



def neighbor_list(quantities, a, cutoff, self_interaction=False,
                  max_nbins=1e6):
    """Compute a neighbor list for an atomic configuration.

    Atoms outside periodic boundaries are mapped into the box. Atoms
    outside nonperiodic boundaries are included in the neighbor list
    but complexity of neighbor list search for those can become n^2.

    The neighbor list is sorted by first atom index 'i', but not by second
    atom index 'j'.

    Parameters:

    quantities: str
        Quantities to compute by the neighbor list algorithm. Each character
        in this string defines a quantity. They are returned in a tuple of
        the same order. Possible quantities are:

           * 'i' : first atom index
           * 'j' : second atom index
           * 'd' : absolute distance
           * 'D' : distance vector
           * 'S' : shift vector (number of cell boundaries crossed by the bond
             between atom i and j). With the shift vector S, the
             distances D between atoms can be computed from:
             D = a.positions[j]-a.positions[i]+S.dot(a.cell)
    a: :class:`ase.Atoms`
        Atomic configuration.
    cutoff: float or dict
        Cutoff for neighbor search. It can be:

            * A single float: This is a global cutoff for all elements.
            * A dictionary: This specifies cutoff values for element
              pairs. Specification accepts element numbers of symbols.
              Example: {(1, 6): 1.1, (1, 1): 1.0, ('C', 'C'): 1.85}
            * A list/array with a per atom value: This specifies the radius of
              an atomic sphere for each atoms. If spheres overlap, atoms are
              within each others neighborhood. See :func:`~ase.neighborlist.natural_cutoffs`
              for an example on how to get such a list.

    self_interaction: bool
        Return the atom itself as its own neighbor if set to true.
        Default: False
    max_nbins: int
        Maximum number of bins used in neighbor search. This is used to limit
        the maximum amount of memory required by the neighbor list.

    Returns:

    i, j, ...: array
        Tuple with arrays for each quantity specified above. Indices in `i`
        are returned in ascending order 0..len(a), but the order of (i,j)
        pairs is not guaranteed.

    Examples:

    Examples assume Atoms object *a* and numpy imported as *np*.

    1. Coordination counting::

        i = neighbor_list('i', a, 1.85)
        coord = np.bincount(i)

    2. Coordination counting with different cutoffs for each pair of species::

        i = neighbor_list('i', a,
                          {('H', 'H'): 1.1, ('C', 'H'): 1.3, ('C', 'C'): 1.85})
        coord = np.bincount(i)

    3. Pair distribution function::

        d = neighbor_list('d', a, 10.00)
        h, bin_edges = np.histogram(d, bins=100)
        pdf = h/(4*np.pi/3*(bin_edges[1:]**3 - bin_edges[:-1]**3)) * a.get_volume()/len(a)

    4. Pair potential::

        i, j, d, D = neighbor_list('ijdD', a, 5.0)
        energy = (-C/d**6).sum()
        pair_forces = (6*C/d**5  * (D/d).T).T
        forces_x = np.bincount(j, weights=pair_forces[:, 0], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 0], minlength=len(a))
        forces_y = np.bincount(j, weights=pair_forces[:, 1], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 1], minlength=len(a))
        forces_z = np.bincount(j, weights=pair_forces[:, 2], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 2], minlength=len(a))

    5. Dynamical matrix for a pair potential stored in a block sparse format::

        from scipy.sparse import bsr_matrix
        i, j, dr, abs_dr = neighbor_list('ijDd', atoms)
        energy = (dr.T / abs_dr).T
        dynmat = -(dde * (energy.reshape(-1, 3, 1) * energy.reshape(-1, 1, 3)).T).T \
                 -(de / abs_dr * (np.eye(3, dtype=energy.dtype) - \
                   (energy.reshape(-1, 3, 1) * energy.reshape(-1, 1, 3))).T).T
        dynmat_bsr = bsr_matrix((dynmat, j, first_i), shape=(3*len(a), 3*len(a)))

        dynmat_diag = np.empty((len(a), 3, 3))
        for x in range(3):
            for y in range(3):
                dynmat_diag[:, x, y] = -np.bincount(i, weights=dynmat[:, x, y])

        dynmat_bsr += bsr_matrix((dynmat_diag, np.arange(len(a)),
                                  np.arange(len(a) + 1)),
                                 shape=(3 * len(a), 3 * len(a)))

    """
    return primitive_neighbor_list(quantities, a.pbc,
                                   a.get_cell(complete=True),
                                   a.positions, cutoff, species=a.info['species'],
                                   self_interaction=self_interaction,
                                   max_nbins=max_nbins)


if __name__ == "__main__":
    from ase.io import read, write
    from ase.build import molecule, bulk
    from ase.visualize import view
    from ase import Atom, Atoms
    atoms = read('docs/source/_static/datas/tio2.cif')
    atoms.info['species'] = atoms.get_chemical_symbols()
    # atoms = read('docs/source/_static/datas/mof-5.cif')
    atoms.info['species'][0] = 'Ti_1'
    cutoff = {('Ti', 'O'): [0.5, 2.5], ('O', 'O'):[0.5, 1.5]}
    i, j, S = neighbor_list('ijS', atoms, cutoff)
    print(i)
    print(j)
    print(S)



"""
Rendering ase atoms objects using blender.

"""
import bpy
from mathutils import Vector, Matrix
import os
import numpy as np

default_render_settings = {
        'transparent': True,  # transparent background
        'resolution_x': 1000,
        'resolution_y': None,  # 
        'lock_camera_to_view': True,
        'lock_light_to_camera': True,
        'camera_loc': [0, 0, 100],  # x, y is the image plane, z is *out* of the screen
        'camera_target': [0, 0, 0], #
        'camera_type': 'ORTHO',  #  ['PERSP', 'ORTHO']
        'ortho_scale': None, #
        'camera_lens': 50,  #
        'fstop': 0.5,
        'ratio': 1,
        'light_loc': [0, 0, 100],
        'light_type': 'SUN', # 'POINT', 'SUN', 'SPOT', 'AREA'
        'light_energy': 5.0,
        'studiolight': 'Default',  # "basic", "outdoor", "paint", "rim", "studio"
        'world': False,
        'engine': 'BLENDER_WORKBENCH', #'BLENDER_EEVEE' #'BLENDER_WORKBENCH'
        'use_motion_blur': False,
        'motion_blur_position': 'START', 
        'motion_blur_steps': 10,
        'motion_blur_shutter': 30.0,
        'frame': 0,
        'functions': [],
        'run_render': True,
        'output': None,
        'animation': False,
        'save_to_blend': False,
        'queue': None,
        'gpu': True,
        'num_samples': 32,
        }

class Render():
    """
    Object to render atomic structure.

    """
    #
    def __init__(self, label, batoms = None, **kwargs):
        self.label = label
        self.batoms = batoms
        self.camera_name = 'camera_%s'%self.label
        self.light_name = 'light_%s'%self.label
        default_render_settings.update(kwargs)
        self.set_parameters(default_render_settings)
        self.clean_default()
        self.set_coll()
        
        if self.world:
            self.set_world()
    def set_parameters(self, parameters):
        for k, v in parameters.items():
            setattr(self, k, v)
    def clean_default(self, camera = False, light = True):
        if 'Cube' in bpy.data.objects:
            bpy.data.objects.remove(bpy.data.objects["Cube"], do_unlink=True)
        if camera and 'Camera' in bpy.data.cameras:
            bpy.data.cameras.remove(bpy.data.cameras['Camera'])
        if light and 'Light' in bpy.data.lights:
            bpy.data.lights.remove(bpy.data.lights['Light'])
    def set_coll(self):
        name = "%s_render"%self.label
        if name not in bpy.data.collections:
            coll = bpy.data.collections.new(name)
            self.scene.collection.children.link(coll)
    @property
    def scene(self):
        return self.get_scene()
    def get_scene(self):
        return bpy.data.scenes['Scene']
    @property
    def coll(self):
        return self.get_coll()
    def get_coll(self):
        name = "%s_render"%self.label
        return bpy.data.collections[name]
    @property
    def engine(self):
        return self.get_engine()
    @engine.setter
    def engine(self, engine):
        self.set_engine(engine)
    def get_engine(self):
        return self.scene.render.engine
    def set_engine(self, engine):
        if engine.upper() == 'EEVEE':
            engine = 'BLENDER_EEVEE'
        elif engine.upper() == 'WORKBENCH':
            engine = 'BLENDER_WORKBENCH'
        self.scene.render.engine = engine.upper()
    @property
    def camera(self):
        return self.get_camera()
    @camera.setter
    def camera(self, camera):
        self.set_camera(camera)
    def get_camera(self):
        return bpy.data.objects[self.camera_name]
    @property
    def light(self):
        return self.get_light()
    @light.setter
    def light(self, light):
        self.set_light(light)
    def get_light(self):
        return bpy.data.objects[self.light_name]
    def set_world(self, color = [0.9, 0.9, 0.9, 1.0]):
        world = self.scene.world
        world.use_nodes = True
        node_tree = world.node_tree
        rgb_node = node_tree.nodes.new(type="ShaderNodeRGB")
        rgb_node.outputs["Color"].default_value = color
        node_tree.nodes["Background"].inputs["Strength"].default_value = 1.0
        node_tree.links.new(rgb_node.outputs["Color"], node_tree.nodes["Background"].inputs["Color"])
    def set_camera(self, camera_type = None, camera_lens = None,):
        '''

        camera_type: str
            * PERSP
            * ORTHO
        camera_lens: float
            
        # todo lock camera to view
        
        '''
        if not camera_type:
            camera_type = self.camera_type
        if not camera_lens:
            camera_lens = self.camera_lens
        if self.camera_name in bpy.data.objects:
            camera = bpy.data.objects[self.camera_name]
        else:
            if self.camera_name in bpy.data.cameras:
                camera_data = bpy.data.cameras[self.camera_name]
            else:
                camera_data = bpy.data.cameras.new(self.camera_name)
            camera = bpy.data.objects.new(self.camera_name, camera_data)
            self.coll.objects.link(camera)
        camera.data.lens = camera_lens
        camera.data.dof.aperture_fstop = self.fstop
        camera.data.type = camera_type
        target = self.camera_target
        if self.ortho_scale and camera_type == 'ORTHO':
            camera.data.ortho_scale = self.ortho_scale
        camera.location = Vector(self.camera_loc)
        self.look_at(camera, target, roll = 0.0)
        bpy.context.scene.camera = camera
    def set_light(self):
        '''

        light_type: str
            * POINT, Omnidirectional point light source.
            * SUN, Constant direction parallel ray light source.
            * SPOT, Directional cone light source.
            * AREA, Directional area light source.
        light_energy: float

        # track to camera

        '''
        # check light exist or not
        if self.light_name in bpy.data.objects:
            light = bpy.data.objects[self.light_name]
        else:
            if self.light_name in bpy.data.lights:
                light_data = bpy.data.lights[self.light_name]
            else:
                light_data = bpy.data.lights.new(self.light_name, type = self.light_type)
            light = bpy.data.objects.new(self.light_name, light_data)
            self.coll.objects.link(light)
        light.data.type = self.light_type
        light.data.energy = self.light_energy
        light.location = Vector(self.camera_loc)
        light.data.use_nodes = True
        light.data.node_tree.nodes['Emission'].inputs['Strength'].default_value = 0.1
        if self.lock_light_to_camera:
            light.constraints.new(type = 'COPY_LOCATION')
            light.constraints["Copy Location"].target = self.camera
            light.constraints.new(type = 'COPY_ROTATION')
            light.constraints["Copy Rotation"].target = self.camera
        else:
            light.location = Vector(self.light_loc)
            self.look_at(light, self.camera_target)
    def look_at(self, obj, target, roll=0):
        """
        Rotate obj to look at target
        """
        if not isinstance(target, Vector):
            target = Vector(target)
        loc = obj.location
        direction = target - loc
        quat = direction.to_track_quat('-Z', 'Y')
        quat = quat.to_matrix().to_4x4()
        rollMatrix = Matrix.Rotation(roll, 4, 'Z')
        loc = loc.to_tuple()
        obj.matrix_world = quat @ rollMatrix
        obj.location = loc
    def motion_blur(self):
        bpy.context.scene.eevee.use_motion_blur = self.use_motion_blur
        bpy.context.scene.eevee.motion_blur_position = self.motion_blur_position
        bpy.context.scene.eevee.motion_blur_steps = self.motion_blur_steps
        bpy.context.scene.eevee.motion_blur_shutter = self.motion_blur_shutter
        bpy.context.scene.cycles.use_motion_blur = self.use_motion_blur
        bpy.context.scene.cycles.motion_blur_position = self.motion_blur_position
        bpy.context.scene.cycles.motion_blur_shutter = self.motion_blur_shutter

    def set_direction(self, direction = (0, 0, 1), margin = None, canvas = None):
        """
        """
        from batoms.tools import get_canvas
        batoms = self.batoms
        atoms = batoms.get_atoms_with_boundary()
        if not margin:
            sizes = [ba.size.max() for ba in batoms.batoms.values()]
            margin = max(sizes) + 0.5
        if canvas is None:
            canvas, canvas1 = get_canvas(atoms = atoms, direction = direction, margin = margin)
        else:
            if isinstance(canvas, (int, float)):
                canvas = np.array([[0, 0, 0], [canvas, canvas, canvas]])
            canvas1 = canvas
        camera_data = batoms.calc_camera_data(canvas, canvas1, direction = direction)
        self.set_parameters(camera_data)
    def run(self, direction = None, canvas = None, **kwargs):
        """
        """
        from batoms.butils import lock_camera_to_view
        if direction:
            self.set_direction(direction, canvas = canvas)
        if self.lock_camera_to_view:
            lock_camera_to_view(True)
        self.set_parameters(kwargs)
        self.set_camera()
        self.set_light()
        if self.use_motion_blur:
            self.motion_blur()
        self.prepare()
        if self.save_to_blend:
            print('saving to {0}.blend'.format(self.output))
            bpy.ops.wm.save_as_mainfile('EXEC_SCREEN', filepath = '{0}.blend'.format(self.output))
        elif self.run_render:
            bpy.ops.render.render(write_still = 1, animation = self.animation)
    def prepare(self, ):
        self.scene.frame_set(self.frame)
        if self.output is None:
            self.output = self.batoms.label
        self.directory = os.path.split(self.output)[0]
        if self.directory and not os.path.exists(self.directory):
            os.makedirs(self.directory)
        self.scene.render.image_settings.file_format = 'PNG'
        self.scene.render.engine = self.engine.upper()
        if self.engine.upper() == 'BLENDER_WORKBENCH':
            bpy.data.scenes['Scene'].display.shading.studio_light = '%s'%self.studiolight
        if self.engine.upper() == 'BLENDER_EEVEE':
            self.scene.eevee.taa_render_samples = self.num_samples
            self.scene.cycles.use_denoising = True
        if self.engine.upper() == 'CYCLES':
            self.scene.cycles.samples = self.num_samples
            self.scene.cycles.use_denoising = True
            if self.gpu:
                self.scene.cycles.device = 'GPU'
                prefs = bpy.context.preferences.addons['cycles'].preferences
                for device in prefs.devices:
                    device.use = True
                self.scene.render.tile_x = 256
                self.scene.render.tile_y = 256
        self.scene.render.film_transparent = self.transparent
        self.scene.render.resolution_x = self.resolution_x
        self.scene.render.resolution_y = int(self.resolution_x*self.ratio)
        self.scene.render.filepath = '{0}'.format(self.output)
    def export(self, filename = 'blender-ase.obj'):
        if filename.split('.')[-1] == 'obj':
            bpy.ops.export_scene.obj(filename)
        if filename.split('.')[-1] == 'x3d':
            bpy.ops.export_scene.x3d(filename)
        if filename.split('.')[-1] == 'xyz':
            self.export_xyz(filename)

    def render_move_camera(self, filename, loc1, loc2, n):
        from batoms.tools import getEquidistantPoints
        locs = getEquidistantPoints(loc1, loc2, n)
        i = 0
        for loc in locs:
            bpy.data.objects['Camera'].location = loc
            self.render(self, filename + '_{0:3d}'.format(i))
            i += 1
    def __repr__(self) -> str:
        s = '-'*60 + '\n'
        s += 'Engine: %s \n'%(self.engine)
        s += 'Camera: %s %s %s \n'%(self.camera_type, self.camera_loc, self.ortho_scale)
        s += 'Light: %s %s \n'%(self.light_type, self.light_energy)
        s += '-'*60 + '\n'
        return s


import pytest
from batoms.batom import Batom
import numpy as np
from batoms.butils import removeAll

def test_batom():
    """
    """
    removeAll()
    positions = np.array([[0, 0, 0], [1.52, 0, 0]])
    h = Batom('h2o', 'H', positions = positions)
    assert isinstance(h, Batom)
    h.translate([0, 0, 2])
    assert np.allclose(h.positions, np.array([[0, 0, 2], [1.52, 0, 2]]))
    # properties
    h.scale = 2
    assert np.allclose(h.scale, np.array([2, 2, 2]))
    #
    h1 = Batom('1 o', 'H_1', [[0, 0, 2], [2, 0, 2]])
    h = h + h1
    h2 = h.copy('h2o', 'H_2')
    #
    assert isinstance(h2, Batom)
    h3=Batom('atom_h2o_H_2')
    assert isinstance(h3, Batom)

def test_positions():
    from batoms.batom import Batom
    import numpy as np
    removeAll()
    positions = np.array([[0, 0, 0], [1.52, 0, 0]])
    h = Batom('h2o', 'H', positions = positions)
    npositions = positions - np.array([0, 0, 5])
    h.positions = npositions


def test_batom_animation():
    from batoms.batom import Batom
    import numpy as np
    removeAll()
    positions = np.array([[0, 0 ,0], [1.52, 0, 0]])
    o = Batom('co2', 'O', positions)
    images = []
    for i in range(10):
        images.append(positions + np.array([i, 0, 0]))
    o.load_frames(images)


if __name__ == '__main__':
    test_batom()
    test_positions()
    test_batom_animation()
    print('\n Batom: All pass! \n')


"""

Wrapper functions for ASE build

"""

from ase import atom, build
from batoms.batoms import Batoms


def molecule(label, symbol, **kwargs):
    atoms = build.molecule(symbol, **kwargs)
    batoms = Batoms(label = label, atoms = atoms)
    return batoms

def bulk(label, symbol, **kwargs):
    atoms = build.bulk(symbol, **kwargs)
    batoms = Batoms(label = label, atoms = atoms)
    return batoms

def surface(label, lattice, indices, layers, **kwargs):
    if isinstance(lattice, Batoms):
        lattice = lattice.atoms
    atoms = build.surface(lattice, indices, layers, **kwargs)
    atoms.info.pop('species', None)
    batoms = Batoms(label = label, atoms = atoms)
    return batoms



"""
Definition of the Batoms class in the batoms package.

"""

import bpy
from ase import Atoms
from batoms.batom import Batom
from batoms.bondsetting import Bondsetting, build_bondlists, calc_bond_data
from batoms.polyhedrasetting import Polyhedrasetting, build_polyhedralists
from batoms.isosurfacesetting import Isosurfacesetting
from batoms.cell import Bcell
from batoms.render import Render   
from batoms.boundary import search_boundary, search_bond
from batoms.bdraw import draw_cell_cylinder, draw_bond_kind, draw_polyhedra_kind, \
                        draw_isosurface
from batoms.butils import object_mode
import numpy as np
from time import time

import logging
logging.basicConfig(
                    format=('%(levelname)-8s '
                            '[%(funcName)-20s]: %(message)s'),
                    level=logging.INFO)

logger = logging.getLogger(__name__)

subcollections = ['atom', 'cell', 'bond', 'polyhedra', 'instancer', 
'instancer_atom', 'volume', 'ghost', 'boundary', 'skin', 'render', 'text']

class Batoms():
    """
    Batoms object

    The Batoms object is a interface to a batoms collection in Blender.
    A batoms collections is organised in the following way, 
    take water molecule as a example:

    * h2o                          # main collection

      * h2o_atom                   # atoms collection
    
        * atom_h2o_H               # atoms object
    
        * atom_h2o_O               # atoms object

      * h2o_bond                   # bonds collection

        * bond_h2o_H               # bond object
    
        * bond_h2o_O               # bond object
    
      * h2o_cell                   # cell collection
    
      * h2o_instancer              # instancer collection
    
        * instancer_atom_h2o_H     # sphere object
    
        * instancer_atom_h2o_O     # sphere object
    
    Then, a Batoms object is linked to this main collection in Blender. 

    Parameters:

    label: str
        Name for the collection in Blender.
    species: dict or list
        Can be a dict with symbols and positions. Examples:
        {
         'O': [[0, 0, 0.40]], 
         'H': [[0, -0.76, -0.2], [0, 0.76, -0.2]]
        }
        Or can be a list of Baom object.
        [Batom('h2o', 'H', ...), Batom('h2o', 'O', ...)]
    atoms: ase.atoms.Atoms object or a list of ase.atoms.Atoms object
        or pymatgen structure object
    model_type: int
        enum in [0, 1, 2, 3], Default value: 0
    pbc: Bool or three Bool
        Periodic boundary conditions. Examples: True,
        False, (True, False, False).  Default value: False.
    cell: 3x3 matrix or length 3 or 6 vector
        Unit cell.
    segments: list
        value should be int, and in [3, 100000]
        segments and ring_count in bpy.ops.mesh.primitive_uv_sphere_add

    boundary:  list 
        search atoms at the boundary

    Examples:

    >>> from batoms import Batoms
    >>> h2o = Batoms(label = 'h2o', species = {'O': [[0, 0, 0.40]], 
    ...             'H': [[0, -0.76, -0.2], [0, 0.76, -0.2]]})
    
    Here is equivalent:
    
    >>> h = Batom(label = 'h2o', species = 'H', 
    ...           positions = [[0, -0.76, -0.2], [0, 0.76, -0.2]])
    >>> o = Batom(label = 'h2o', species = 'O', 
    ...           positions = [[0, 0, 0.40]])
    >>> h2o = Batoms('h2o', [h, o])

    """
    

    def __init__(self, label = None,
                species = None,
                atoms = None, 
                pbc = False, cell = None,
                bondsetting = None,
                polyhedrasetting = None,
                render = None,
                model_type = 0, 
                polyhedra_type = 0, 
                boundary = [0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
                show_unit_cell = True,
                volume = None,
                segments = [32, 16],
                shape = 'UV_SPHERE',
                kind_props = {},
                color_style = 'JMOL',
                material_style = 'default',
                bsdf_inputs = None,
                movie = False,
                draw = True, 
                 ):
        #
        self.parent = None
        self.bondsetting = bondsetting
        self.polyhedrasetting = polyhedrasetting
        self.render = render
        self.segments = segments
        self.shape = shape
        self.kind_props = kind_props
        self.label = label
        self.color_style = color_style
        self.material_style = material_style
        self.bsdf_inputs = bsdf_inputs
        if species:
            if not self.label:
                self.label = ''.join(['%s%s'%(species, len(positions)) 
                             for sp, positions in species.items()])
            self.set_collection(model_type, polyhedra_type, boundary)
            self.from_species(species, pbc, cell)
        elif atoms:
            if isinstance(atoms, list):
                self.images = atoms
                atoms = self.images[0]
            self.set_collection(model_type, polyhedra_type, boundary)
            if 'ase' in str(type(atoms)):
                self.from_ase(atoms)
            elif 'pymatgen' in str(type(atoms)):
                self.from_pymatgen(atoms)
        elif self.label:
            print('Build from collection')
            self.from_collection(self.label)
        else:
            raise Exception("Failed, species, atoms or coll  \
                  should be provided!"%self.label)
        if not self.bondsetting:
            self.bondsetting = Bondsetting(self.label)
        if not self.polyhedrasetting:
            self.polyhedrasetting = Polyhedrasetting(self.label)
        self.isosurfacesetting = Isosurfacesetting(self.label, volume = volume)
        self.coll.batoms.show_unit_cell = show_unit_cell
        if not self.render:
            self.render = Render(self.label, batoms = self)
        if draw:
            self.draw()
        if movie:
            self.load_frames()
        self.show_index()
    def from_species(self, species, pbc = None, cell = None):
        """
        """
        if isinstance(species, dict):
            for sp, positions in species.items():
                if sp not in self.kind_props: self.kind_props[sp] = {}
                ba = Batom(self.label, sp, positions, segments = self.segments, 
                            shape = self.shape, props = self.kind_props[sp], 
                            material_style=self.material_style, 
                            bsdf_inputs=self.bsdf_inputs, color_style=self.color_style)
                self.coll.children['%s_atom'%self.label].objects.link(ba.batom)
                self.coll.children['%s_instancer'%self.label].objects.link(ba.instancer)
        elif isinstance(species, list):
            for batom in species:
                if not isinstance(batom, Batom):
                    raise Exception('%s is not a Batom object.'%batom)
                self.coll.children['%s_atom'%self.label].objects.link(batom.batom)
                self.coll.children['%s_instancer'%self.label].objects.link(batom.instancer)
        self._cell = Bcell(self.label, cell)
        self.coll.children['%s_cell'%self.label].objects.link(self._cell.bcell)
        self.set_pbc(pbc)
    def from_ase(self, atoms):
        """
        Import structure from ASE atoms.
        """
        if 'species' not in atoms.info:
            atoms.info['species'] = atoms.get_chemical_symbols()
        species_list = list(set(atoms.info['species']))
        for species in species_list:
            indices = [index for index, x in enumerate(atoms.info['species']) 
                        if x == species]
            if species not in self.kind_props: self.kind_props[species] = {}
            ba = Batom(self.label, species, atoms.positions[indices], 
                        segments = self.segments, shape = self.shape, 
                        props = self.kind_props[species], 
                        material_style=self.material_style, 
                        bsdf_inputs=self.bsdf_inputs, 
                        color_style=self.color_style)
            self.coll.children['%s_atom'%self.label].objects.link(ba.batom)
            self.coll.children['%s_instancer'%self.label].objects.link(ba.instancer)
        self.coll.batoms.pbc = self.npbool2bool(atoms.pbc)
        self._cell = Bcell(self.label, atoms.cell)
        self.coll.children['%s_cell'%self.label].objects.link(self._cell.bcell)
    def from_pymatgen(self, structure):
        """
        Import structure from Pymatgen structure.
        """
        symbols = [str(site.specie.symbol) for site in structure]
        if hasattr(structure, "lattice"):
            cell = structure.lattice.matrix
            pbc = True
        else:
            cell = None
            pbc = False
        species_list = list(set(symbols))
        for species in species_list:
            positions = [structure[index].coords for index, x 
                           in enumerate(symbols) if x == species]
            ba = Batom(self.label, species, positions, segments = self.segments, 
                       shape = self.shape, material_style=self.material_style, 
                       bsdf_inputs=self.bsdf_inputs, color_style=self.color_style)
            self.coll.children['%s_atom'%self.label].objects.link(ba.batom)
            self.coll.children['%s_instancer'%self.label].objects.link(ba.instancer)
        self.set_pbc(pbc)
        self._cell = Bcell(self.label, cell)
        self.coll.children['%s_cell'%self.label].objects.link(self._cell.bcell)
    def from_collection(self, collection_name):
        """
        """
        if collection_name not in bpy.data.collections:
            raise Exception("%s is not a collection!"%collection_name)
        elif not bpy.data.collections[collection_name].batoms.is_batoms:
            raise Exception("%s is not Batoms collection!"%collection_name)
        self.label = collection_name
        self._cell = Bcell(label = collection_name)
        self.bondsetting = Bondsetting(self.label)

    def npbool2bool(self, pbc):
        """
        """
        newpbc = []
        for i in range(3):
            if pbc[i]:
                newpbc.append(True)
            else:
                newpbc.append(False)
        return newpbc
    def set_collection(self, model_type = 0, polyhedra_type = 0, boundary = [0, 0, 0]):
        """
        build main collection and its child collections.
        """
        for coll in bpy.data.collections:
            if self.label == coll.name:
                raise Exception("Failed, the name %s already in use!"%self.label)
        coll = bpy.data.collections.new(self.label)
        self.coll.batoms.is_batoms = True
        self.scene.collection.children.link(self.coll)
        for sub_name in subcollections:
            subcoll = bpy.data.collections.new('%s_%s'%(self.label, sub_name))
            self.coll.children.link(subcoll)
        self.coll.batoms.model_type = str(model_type)
        self.coll.batoms.polyhedra_type = str(polyhedra_type)
        self.coll.batoms.boundary = boundary
    
    def draw_cell(self, celllinewidth = 0.03):
        """
        Draw unit cell
        """
        object_mode()
        cell_vertices = self.cell.verts
        if np.max(abs(cell_vertices)) < 1e-6:
            return 0
        self.clean_atoms_objects('cell', ['cylinder', 'point'])
        if self.show_unit_cell:
            draw_cell_cylinder(self.coll.children['%s_cell'%self.label], 
                            cell_vertices, 
                            label = self.label, 
                            celllinewidth = celllinewidth)
    def draw_bonds(self):
        """
        Draw bonds.

        Parameters:

        cutoff: float
            cutoff used to build bond pairs.
        """
        # if not self.bondlist:
        object_mode()
        self.clean_atoms_objects('bond')
        atoms = self.get_atoms_with_boundary()
        self.bondlist = build_bondlists(atoms, self.bondsetting.cutoff_dict)
        bond_kinds = calc_bond_data(self, self.bondlist, self.bondsetting)
        for species, bond_data in bond_kinds.items():
            draw_bond_kind(species, bond_data, label = self.label, 
                        coll = self.coll.children['%s_bond'%self.label])
    def draw_polyhedras(self, bondlist = None, show_edge = True):
        """
        Draw bonds.
        Parameters:

        cutoff: float
            cutoff used to build bond pairs.
        """
        object_mode()
        self.clean_atoms_objects('polyhedra')
        atoms = self.get_atoms_with_boundary(X = True)
        if bondlist is None:
            bondlist = build_bondlists(atoms, self.bondsetting.cutoff_dict)
        polyhedra_kinds = build_polyhedralists(atoms, bondlist, 
                          self.bondsetting, self.polyhedrasetting)
        for species, polyhedra_data in polyhedra_kinds.items():
            draw_polyhedra_kind(species, polyhedra_data, label = self.label,
                        coll = self.coll.children['%s_polyhedra'%self.label], show_edge = show_edge)
    def draw_isosurface(self):
        """
        Draw bonds.

        Parameters:

        isosurface: list
            isosurface data.
        """
        object_mode()
        self.clean_atoms_objects('volume', ['isosurface'])
        isosurface = self.isosurfacesetting.build_isosurface(self.cell)
        for verts, faces, color in isosurface:
            draw_isosurface(self.coll.children['%s_volume'%self.label],
                            verts, faces, color = color)
    def draw_cavity_sphere(self, radius, boundary = [[0, 1], [0, 1], [0, 1]]):
        """
        cavity
        for porous materials
        >>> from ase.io import read
        >>> atoms = read('docs/source/_static/datas/mof-5.cif')
        """
        from batoms.tools import find_cage_sphere
        object_mode()
        self.clean_atoms_objects('ghost')
        positions = find_cage_sphere(self.cell, self.atoms.positions, radius, boundary = boundary)
        ba = Batom(self.label, 'X', positions, scale = radius*0.9, 
                   material_style='default', bsdf_inputs=self.bsdf_inputs, 
                   color_style=self.color_style)
        
        # ba.color = [ba.color[0], ba.color[1], ba.color[2], 0.8]
        self.coll.children['%s_ghost'%self.label].objects.link(ba.batom)
        self.coll.children['%s_ghost'%self.label].objects.link(ba.instancer)
    def draw_cavity(self, bondlists = None):
        """
        cavity
        for porous materials
        >>> from ase.io import read
        >>> atoms = read('docs/source/_static/datas/mof-5.cif')
        """
        from batoms.tools import find_cage
        from batoms.bdraw import draw_cavity
        object_mode()
        self.clean_atoms_objects('ghost')
        atoms = self.get_atoms_with_boundary()
        positions = atoms.positions
        if bondlists is None:
            bondlists = build_bondlists(atoms, self.bondsetting.cutoff_dict)
        # calc data
        vertices = []
        edges = []
        # offset = bondlists[:, 2:5]
        # R = np.dot(offset, atoms.cell)
        # pos1 = positions[bondlists[:, 0]]
        # pos2 = positions[bondlists[:, 1]] + R
        # n = len(pos1)
        # vertices = np.append(pos1, pos2, axis = 0)
        # index1 = np.array(range(n)).reshape(-1, 1)
        # index2 = np.array(range(n)).reshape(-1, 1) + n
        # edges = np.append(index1, index2, axis = 1)
        edges = bondlists[:, [0, 1]]
        edges.astype(int)
        edges = edges.tolist()
        # draw edge based on bonds
        kind = 'cavity'
        datas = {'vertices': positions, 'edges': edges, 'color': [0.4, 0.4, 0.1], 'transmit': 1.0}
        draw_cavity(kind, datas, self.label, coll = self.coll.children['%s_polyhedra'%self.label])
        
    def clean_atoms_objects(self, coll, names = None):
        """
        remove all bond object in the bond collection
        """
        if not names:
            for obj in self.coll.children['%s_%s'%(self.label, coll)].all_objects:
                bpy.data.objects.remove(obj, do_unlink = True)
        else:
            for name in names:
                for obj in self.coll.children['%s_%s'%(self.label, coll)].all_objects:
                    if name in obj.name:
                        bpy.data.objects.remove(obj, do_unlink = True)
                
    def show_index(self, index_type = 0):
        """
        """
        bpy.context.preferences.view.show_developer_ui = True
        for a in bpy.context.screen.areas:
            if a.type == 'VIEW_3D':
                overlay = a.spaces.active.overlay
                overlay.show_extra_indices = True
    @property
    def scene(self):
        return self.get_scene()
    def get_scene(self):
        return bpy.data.scenes['Scene']
    @property
    def scale(self):
        return self.get_scale()
    @scale.setter
    def scale(self, scale):
        self.set_scale(scale)
    def get_scale(self):
        scale = {}
        for coll in [self.batoms, self.batoms_boundary, self.batoms_skin]:
            for batom in coll.values():
                scale[batom.species] = batom.scale
        return scale
    def set_scale(self, scale):
        for coll in [self.batoms, self.batoms_boundary, self.batoms_skin]:
            for batom in coll.values():
                batom.scale = scale
    def draw(self, model_type = None):
        """
        Draw atoms, bonds, polyhedra.

        Parameters:

        model_type: str

        """
        if model_type is not None and model_type not in [0, 1, 2, 3]:
            raise Exception('model_type %s should be: 0, 1, 2, 3'%model_type)
        if not model_type:
            model_type = self.model_type
        else:
            self.model_type = model_type
        # self.draw_cell()
        bpy.ops.ed.undo_push()
        self.clean_atoms_objects('bond')
        bpy.ops.ed.undo_push()
        self.clean_atoms_objects('polyhedra')
        bpy.ops.ed.undo_push()
        if model_type == 0:
            self.scale = 1.0
        elif model_type == 1:
            self.scale = 0.4
            self.draw_bonds()
        elif model_type == 2:
            if self.polyhedra_type == 0:
                self.scale = 0.4
                self.draw_bonds()
                self.draw_polyhedras(self.bondlist)
            if self.polyhedra_type == 1:
                self.scale = 0.4
                self.draw_polyhedras()
            elif self.polyhedra_type == 2:
                self.scale = 0.01
                for b in self.bondsetting:
                    if b.polyhedra:
                        self.batoms[b.symbol1].scale = 0.4
                self.draw_polyhedras()
            elif self.polyhedra_type == 3:
                self.scale = 0.01
                self.draw_polyhedras()
        elif model_type == 3:
            self.scale = 0.01
            self.draw_bonds()
        if self.isosurfacesetting.npoint > 0:
            self.draw_isosurface()
    def replace(self, species1, species2, index = []):
        """
        replace atoms.

        Parameters:
        
        index: list
            index of atoms will be replaced.

        species1: str

        species2: str
            atoms will be changed to this element.

        >>> from ase.build import molecule, fcc111
        >>> from batoms.batoms import Batoms
        >>> pt111 = fcc111('Pt', (5, 5, 4), vacuum = 5.0)
        >>> pt111 = Batoms(atoms = pt111, label = 'pt111')
        >>> pt111.replace('Pt', 'Au', [93])
        >>> pt111.replace('Pt', 'Au', range(20))

        """
        # if kind exists, merger, otherwise build a new kind and add.
        object_mode()
        positions = self.batoms[species1].positions[index]
        if species2 in self.batoms:
            self.batoms[species2].add_vertices(positions)
        else:
            ba = Batom(self.label, species2, positions, segments = self.segments, 
                       shape = self.shape, material_style=self.material_style, 
                       bsdf_inputs=self.bsdf_inputs, color_style=self.color_style)
            self.coll.children['%s_atom'%self.label].objects.link(ba.batom)
            self.coll.children['%s_instancer'%self.label].objects.link(ba.instancer)
            for sp in self.species:
                self.bondsetting.add_bonds([[species2, sp]])
            self.polyhedrasetting.add_polyhedras([sp])
        self.batoms[species1].delete(index)
            
    
    def delete(self, species, index = []):
        """
        delete atoms.

        species: str

        index: list
            index of atoms to be delete
        
        For example, delete the second atom in H species.
        Please note that index start from 0.

        >>> h2o.delete([1])

        """
        self.batoms[species].delete(index)
    def translate(self, displacement):
        """Translate atomic positions.

        The displacement argument is an xyz vector.

        For example, move h2o molecule by a vector [0, 0, 5]

        >>> h2o.translate([0, 0, 5])

        """
        object_mode()
        bpy.ops.object.select_all(action='DESELECT')
        for obj in self.coll.all_objects:
            if 'instancer' not in obj.name and 'boundary' not in obj.name:
                obj.select_set(True)
        bpy.ops.transform.translate(value=displacement)
    def rotate(self, angle, axis = 'Z', orient_type = 'GLOBAL'):
        """Rotate atomic based on a axis and an angle.

        Parameters:

        angle: float
            Angle that the atoms is rotated around the axis.
        axis: str
            'X', 'Y' or 'Z'.

        For example, rotate h2o molecule 90 degree around 'Z' axis:
        
        >>> h2o.rotate(90, 'Z')

        """
        object_mode()
        bpy.ops.object.select_all(action='DESELECT')
        for coll in self.coll.children:
            for obj in coll.objects:
                obj.select_set(True)
        bpy.ops.transform.rotate(value=angle, orient_axis=axis.upper(), 
                                 orient_type = orient_type)
    
    def __getitem__(self, species):
        """Return a subset of the Batom.

        species -- str, describing which batom to return.
        """

        if isinstance(species, str):
            if species not in self.batoms:
                raise SystemExit('%s is not in this structure'%species)
            return self.batoms[species]
        elif isinstance(species, list):
            raise SystemExit('dict not supported yet!')
    def __len__(self):
        return len(self.positions)
    def __repr__(self) -> str:
        text = []
        text.append('label={0}, '.format(self.label))
        text.append('species='.format(self.cell))
        text.append('%s '%(list(self.batoms)))
        text.append('cell={0}, '.format(self.cell))
        text.append('pbc={0}'.format(self.pbc))
        text = "".join(text)
        text = "Batoms(%s)"%text
        return text
    def __add__(self, other):
        self += other
        return self
    def __iadd__(self, other):
        self.extend(other)
        return self
    def extend(self, other):
        """
        Extend batoms object by appending batoms from *other*.
        
        >>> from ase.build import molecule, fcc111
        >>> from batoms.batoms import Batoms
        >>> import numpy as np
        >>> co = molecule('CO')
        >>> co = Batoms(atoms = co, draw = True)
        >>> au = fcc111('Au', (5, 5, 4), vacuum=5.0)
        >>> au = Batoms(atoms = au, draw = True)
        >>> co.translate(au.atoms[-1].position + np.array([0, 0, 2]))
        >>> au.extend(co)
        >>> au.write('au111-co.cif')
        
        or,

        >>> au = au + co

        """
        from batoms.butils import remove_collection
        # bond first
        self.bondsetting.extend(other.bondsetting)
        self.polyhedrasetting.extend(other.polyhedrasetting)
        # atom
        for species, batom in other.batoms.items():
            if species in self.species:
                self.batoms[species].extend(batom)
            else:
                ba = batom.copy(self.label, species)
                t = self.cell.location - ba.location
                ba.batom.location = self.cell.location
                bpy.context.view_layer.update()
                ba.positions = ba.positions - t
                self.coll.children['%s_atom'%self.label].objects.link(ba.batom)
        remove_collection(other.label)
    def __imul__(self, m):
        """
        """
        for species, batom in self.batoms.items():
            batom.repeat(m, self.cell)
        self.cell.repeat(m)
        self.draw()
    def repeat(self, rep):
        """
        Create new repeated atoms object.

        >>> from ase.build import bulk
        >>> from batoms.batoms import Batoms
        >>> au = bulk('Au', cubic = True)
        >>> au = Batoms(atoms = au)
        >>> au.draw()
        >>> au.repeat([2, 2, 2])
        
        """
        self.__imul__(rep)
    def __mul__(self, rep):
        self.repeat(rep)
        return self
    def copy(self, label = None):
        """
        Return a copy.

        name: str
            The name of the copy.

        For example, copy h2o molecule:
        
        >>> h2o_new = h2o.copy(label = 'h2o_new')

        """
        if not label:
            label = self.label + 'copy'
        species_dict = {x:self.batoms[x].copy(label, x) for x in self.species}
        batoms = self.__class__(species = species_dict, label = label, 
                                cell = self.cell.verts, pbc = self.pbc, 
                                model_type = self.coll.batoms.model_type)
        batoms.translate([2, 2, 2])
        batoms.bondsetting = self.bondsetting.copy(label)
        batoms.polyhedrasetting = self.polyhedrasetting.copy(label)
        return batoms
    def write(self, filename, local = True):
        """
        Save atoms to file.

        >>> h2o.write('h2o.cif')
        
        """
        atoms = self.batoms2atoms(self.batoms, local = local)
        atoms.write(filename)
    def update(self):
        """
        """
        pass
    @property
    def coll(self):
        return self.get_coll()
    def get_coll(self):
        return bpy.data.collections[self.label]
    @property
    def coll_atom(self):
        return self.get_coll_atom()
    def get_coll_atom(self):
        return self.coll.children['%s_atom'%self.label]
    @property
    def coll_boundary(self):
        return self.get_coll_boundary()
    def get_coll_boundary(self):
        return self.coll.children['%s_boundary'%self.label]
    @property
    def coll_skin(self):
        return self.get_coll_skin()
    def get_coll_skin(self):
        return self.coll.children['%s_skin'%self.label]
    @property
    def cell(self):
        return self._cell
    @cell.setter
    def cell(self, cell):
        from ase.cell import Cell
        cell = Cell.ascell(cell)
        self._cell[:] = cell
    def set_cell(self, cell, scale_atoms=False):
        """Set unit cell vectors.

        Parameters:

        Examples:

        """
        from ase.cell import Cell
        from ase.geometry.cell import complete_cell

        cell = Cell.new(cell)
        oldcell = Cell(self.cell)
        self.cell = cell
        if scale_atoms:
            M = np.linalg.solve(oldcell.complete(), cell.complete())
            for ba in self.batoms.values():
                ba.positions = np.dot(ba.positions(), M)
    @property
    def pbc(self):
        return self.get_pbc()
    @pbc.setter
    def pbc(self, pbc):
        self.set_pbc(pbc)
    def get_pbc(self):
        return list(self.coll.batoms.pbc)
    def set_pbc(self, pbc):
        if isinstance(pbc, bool):
            pbc = [pbc]*3
        self.coll.batoms.pbc = pbc
    @property
    def boundary(self):
        return self.get_boundary()
    @boundary.setter
    def boundary(self, boundary):
        if boundary is not None:
            if isinstance(boundary, (int, float)):
                boundary = np.array([-boundary, 1 + boundary]*3)
            elif len(boundary) == 3:
                if isinstance(boundary[0], (int, float)):
                    boundary = np.array([[-boundary[0], 1 + boundary[0]],
                                      [-boundary[1], 1 + boundary[1]],
                                      [-boundary[2], 1 + boundary[2]]])
                elif len(boundary[0]) == 2:
                    boundary = np.array(boundary)
            else:
                raise Exception('Wrong boundary setting!')
            self.coll.batoms.boundary = boundary[:].flatten()
        self.update_boundary()
    def get_boundary(self):
        boundary = np.array(self.coll.batoms.boundary)
        return boundary.reshape(3, -1)
    def update_boundary(self):
        """
        >>> from batoms.batoms import Batoms
        >>> from ase.io import read
        >>> atoms = read('docs/source/_static/datas/tio2.cif')
        >>> tio2 = Batoms(label = 'tio2', atoms = atoms, model_type = '2', 
            polyhedra_dict = {'Ti': ['O']}, color_style="VESTA")
        >>> tio2.boundary = 0.4
        """
        tstart = time()
        boundary = self.boundary
        atoms0 = self.atoms
        if atoms0.pbc.any():
            # find boudary atoms
            atoms_boundary, offsets_skin = search_boundary(atoms0, boundary, self.bondsetting.maxcutoff)
            self.draw_boundary_atoms(atoms_boundary)
            # search bond
            # include the boundary atoms
            bondlists = build_bondlists(atoms0, self.bondsetting.cutoff_dict)
            offsets_skin1, bondlist1, offsets_skin2, bondlist2 = \
                    self.bondsetting.search_bond_list(atoms0, bondlists, offsets_skin)
            # search type 1
            offset_skin_1 = search_bond(atoms0.get_scaled_positions(), offsets_skin1, bondlist1, boundary)
            # search type 2
            offset_skin_2 = search_bond(atoms0.get_scaled_positions(), offsets_skin2, bondlist2, boundary, recursive=True)
            # search type 1 final
            offset_skin_3 = search_bond(atoms0.get_scaled_positions(), offset_skin_2, bondlist1, boundary)
            self.draw_search_bond_atoms(atoms0, offset_skin_1, offset_skin_2, offset_skin_3)
            # print('search skin: {0:10.2f} s'.format(time() - tstart))
    def draw_boundary_atoms(self, atoms_boundary):
        self.clean_atoms_objects('boundary')
        if len(atoms_boundary) == 0: return 0
        species = set(atoms_boundary.info['species'])
        for sp in species:
            positions = atoms_boundary[atoms_boundary.info['species']==sp].positions
            positions = positions + self.batoms[sp].location
            ba = Batom(self.label, '%s_boundary'%(sp), positions, scale = self.batoms[sp].scale, 
                        segments = self.segments, shape = self.shape, material=self.batoms[sp].material)
            self.coll.children['%s_boundary'%self.label].objects.link(ba.batom)
    def draw_search_bond_atoms(self, atoms0, offsets_search_1, offsets_search_2, offsets_search_3):
        # print(atoms)
        self.clean_atoms_objects('skin')
        offsets_search = np.append(offsets_search_1, offsets_search_2, axis = 0)
        offsets_search = np.append(offsets_search, offsets_search_3, axis = 0)
        if len(offsets_search) == 0: return 0
        offsets_search = offsets_search.astype(int)
        atoms_search_bond = atoms0[offsets_search[:, 0]]
        atoms_search_bond.positions = atoms_search_bond.positions + np.dot(offsets_search[:, 1:], atoms0.cell)
        atoms_search_bond.info['species'] = np.array(atoms0.info['species'])[offsets_search[:, 0]]
        species = set(atoms_search_bond.info['species'])
        for sp in species:
            positions = atoms_search_bond[atoms_search_bond.info['species']==sp].positions
            positions = positions + self.batoms[sp].location
            ba = Batom(self.label, '%s_skin'%(sp), positions, scale = self.batoms[sp].scale, 
                        segments = self.segments, shape = self.shape, material=self.batoms[sp].material)
            self.coll.children['%s_skin'%self.label].objects.link(ba.batom)
        # print('update skin: {0:10.2f} s'.format(time() - tstart))

    @property
    def model_type(self):
        return self.get_model_type()
    @model_type.setter
    def model_type(self, model_type):
        self.set_model_type(model_type)
    def get_model_type(self):
        return int(self.coll.batoms.model_type)
    def set_model_type(self, model_type):
        self.coll.batoms.model_type = str(model_type)
        self.draw()
    @property
    def polyhedra_type(self):
        return self.get_polyhedra_type()
    @polyhedra_type.setter
    def polyhedra_type(self, polyhedra_type):
        self.set_polyhedra_type(polyhedra_type)
    def get_polyhedra_type(self):
        return int(self.coll.batoms.polyhedra_type)
    def set_polyhedra_type(self, polyhedra_type):
        self.coll.batoms.polyhedra_type = str(polyhedra_type)
        self.draw()
    @property
    def show_unit_cell(self):
        return self.get_show_unit_cell()
    @show_unit_cell.setter
    def show_unit_cell(self, show_unit_cell):
        self.set_show_unit_cell(show_unit_cell)
    def get_show_unit_cell(self):
        return self.coll.batoms.show_unit_cell
    def set_show_unit_cell(self, show_unit_cell):
        self.coll.batoms.show_unit_cell = show_unit_cell
        self.draw_cell()
    def get_atoms(self, batoms):
        return self.batoms2atoms(batoms)
    @property
    def atoms(self):
        return self.get_atoms(self.batoms)
    @property
    def atoms_boundary(self):
        return self.get_atoms(self.batoms_boundary)
    @property
    def atoms_skin(self):
        return self.get_atoms(self.batoms_skin)
    def get_atoms_with_boundary(self, X = False):
        """
        build ASE atoms from batoms dict.
        """
        atoms = self.batoms2atoms(self.batoms, X = X)
        atoms_boundary = self.batoms2atoms(self.batoms_boundary, X = X)
        atoms_skin = self.batoms2atoms(self.batoms_skin, X = X)
        species = atoms.info['species'] + atoms_boundary.info['species'] + atoms_skin.info['species']
        atoms = atoms + atoms_boundary + atoms_skin
        atoms.info['species'] = species
        atoms.pbc = False
        return atoms
    @property
    def positions(self):
        return self.get_positions()
    def get_positions(self):
        return self.atoms.positions
    def get_scaled_positions(self):
        return self.atoms.get_scaled_positions()
    @property
    def species(self):
        return self.get_species()
    def get_species(self):
        """
        build species from collection.
        """
        species = []
        for ba in self.coll_atom.objects:
            species.append(ba.batom.species)
        return species
    @property
    def batoms(self):
        return self.get_batoms()
    def get_batoms(self):
        batoms = {}
        for ba in self.coll_atom.objects:
            batoms[ba.batom.species] = Batom(ba.name)
        return batoms
    @property
    def batoms_boundary(self):
        return self.get_batoms_boundary()
    def get_batoms_boundary(self):
        batoms_boundary = {}
        for ba in self.coll_boundary.objects:
            batoms_boundary[ba.batom.species] = Batom(ba.name)
        return batoms_boundary
    @property
    def batoms_skin(self):
        return self.get_batoms_skin()
    def get_batoms_skin(self):
        batoms_skin = {}
        for ba in self.coll_skin.objects:
            batoms_skin[ba.batom.species] = Batom(ba.name)
        return batoms_skin
    def batoms2atoms(self, batoms, local = False, X = False):
        object_mode()
        atoms = Atoms()
        species_list = []
        symbols = []
        positions = []
        for species, batom in batoms.items():
            # ghost site will not save 
            if not X and batom.element == 'X': continue
            if species[-9:] == '_boundary': species = species[0:-9]
            if species[-5:] == '_skin': species = species[0:-5]
            species_list.extend([species]*len(batom))
            symbol = [batom.element]*len(batom)
            symbols.extend(symbol)
            if local:
                positions.extend(batom.local_positions)
            else:
                positions.extend(batom.positions)
        atoms = Atoms(symbols, positions, cell = self.cell, pbc = self.pbc)
        atoms.info['species'] = species_list
        return atoms
    def draw_constraints(self):
        """
        """
        #
        constr = self.atoms.constraints
        self.constrainatoms = []
        for c in constr:
            if isinstance(c, FixAtoms):
                for n, i in enumerate(c.index):
                    self.constrainatoms += [i]
    
    def highlight_atoms(self, indexs, shape = 'sphere', radius_scale=1.4,
                           color=(0.0, 1.0, 0.0), transmit=0.6):
        """
        """
        object_mode()
        for index in indexs:
            loc = self.positions[index]
            ele = self.symbols[index]
            radii = radius_scale * self.atom_kinds[ele]['radius']
            if shape == 'cube':
                bpy.ops.mesh.primitive_cube_add(location=loc, size=radii*2)
            else:
                bpy.ops.mesh.primitive_uv_sphere_add(location=loc, radius=radii)
            ball = bpy.context.view_layer.objects.active
            bpy.ops.object.shade_smooth()
            ball.data.materials.append(material)
            ball.show_transparent = True
            self.coll_highlight.objects.link(ball)
    def load_frames(self, images = None):
        """

        images: list
            list of atoms. All atoms show have same species and length.
            
        >>> from ase.io import read
        >>> from batoms import Batoms
        >>> images = read('h2o-animation.xyz', index = ':')
        >>> h2o = Batoms(label = 'h2o', atoms = images)
        >>> h2o.load_frames()
        >>> h2o.render(animation = True)

        """
        if not images:
            images = self.images
        if len(self.atoms) != len(images[0]):
            raise Exception("Number of atoms %s is not equal to %s."%(len(self.atoms), len(images[0])))
        atoms = images[0]
        if 'species' not in atoms.info:
            atoms.info['species'] = atoms.get_chemical_symbols()
        positions = np.array([atoms.positions for atoms in images])
        for species, ba in self.batoms.items():
            index = [atom.index for atom in atoms if atoms.info['species'][atom.index] == species]
            ba.load_frames(positions[:, index])
    
    def calc_camera_data(self, canvas, canvas1, direction = (0, 0, 1)):
        """
        """
        from scipy.spatial.transform import Rotation as R
        camera_target = np.mean(canvas, axis=0)
        camera_data = {}
        width = canvas1[1, 0] - canvas1[0, 0]
        height = canvas1[1, 1] - canvas1[0, 1]
        depth = canvas1[1, 1] - canvas1[0, 1]
        ortho_scale = max(width, height)
        #
        direction = direction/np.linalg.norm(direction)
        location = camera_target + direction*depth
        camera_data = {'camera_loc': location, 'camera_target': camera_target,
                        'ortho_scale': ortho_scale, 'ratio': height/width}
        return camera_data  

    def get_distances(self, species1, i, species2, indices, mic=False):
        """
        Return distances of atom No.i with a list of atoms.

        Use mic=True to use the Minimum Image Convention.

        >>> h2o.get_distances('O', 0, 'H', [0, 1])
        """
        from ase.geometry import get_distances

        p1 = self.batoms[species1][i]
        p2 = self.batoms[species2][indices]
        cell = None
        pbc = None
        if mic:
            cell = self.cell
            pbc = self.pbc
        D, D_len = get_distances(p1, p2, cell=cell, pbc=pbc)
        D_len.shape = (-1,)
        return D_len
    def get_angle(self, species1, i1, species2, i2, species3, i3, mic=False):
        """
        Get angle in degrees between the vectors i2->i1 and
        i2->i3.
        Use mic=True to use the Minimum Image Convention and calculate the
        angle across periodic boundaries.

        >>> h2o.get_angle('H', 0, 'O', 0, 'H', 1)

        """
        from ase.geometry import get_angles
        p1 = self.batoms[species1][i1]
        p2 = self.batoms[species2][i2]
        p3 = self.batoms[species3][i3]
        v12 = p1 - p2
        v32 = p3 - p2
        cell = None
        pbc = None
        if mic:
            cell = self.cell
            pbc = self.pbc
        return get_angles([v12], [v32], cell=cell, pbc=pbc)
    def get_center_of_mass(self, scaled=False):
        """Get the center of mass.

        If scaled=True the center of mass in scaled coordinates
        is returned."""
        return self.atoms.get_center_of_mass(scaled = scaled)
